/*
 * except_i386.c
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * 㳰ϥɥ
 */


#include <sys/types.h>
#include <sys/param.h>
#include <kern/lib.h>
#include <kern/proc.h>
#include <kern/signal.h>
#include <kern/mm.h>
#include <kern/lock.h>
#include <machine/segment.h>
#include <machine/lib.h>
#include <machine/fpu.h>
#include <machine/mode_switch.h>
#include <machine/interrupt.h>
#include <machine/except.h>

#include <kern/debug.h>


/************************************************************************************************
 *
 * 㳰ϥɥ
 *
 ************************************************************************************************/

//================================== PRIVATE ============================================

/*
 * 㳰ϥɥ
 */
static void exception(const char *message, int signal, INTERRUPT_FRAME *frame)
{
	printk("\nException! %s.",message);
	printContext(frame);
//**************************************
if ((frame->cs == KERNEL_CODE_DES) || (frame->cs == APM_CODE_DES)) {
	idle();
}
//**************************************
	forceSendSignal(get_current_task(), signal, TASK_SIGNAL_WAIT);
}

/*
 * 㳰ϥɥ
 * åѤޤ줿ͤɽƥɥ˥פ
 */
void tmp_except(const char *message, INTERRUPT_FRAME *frame)
{
	printk("\nException! %s.\n",message);
	if((frame->cs==KERNEL_CODE_DES)||(frame->cs==APM_CODE_DES)){
		printk("error=0x%x eip=0x%x cs=0x%x esp=0x%x proc=0x%x cpu=%d\n",
			frame->error,frame->eip,frame->cs,&frame->user_esp, get_current_task(), getCpu());
	}
	else{
		printk("error=0x%x eip=0x%x cs=0x%x esp=0x%x,ss=0x%x\n",frame->error,frame->eip,frame->cs,frame->user_esp,frame->user_ss);
	}

	idle();
}

//================================== PUBLIC =============================================

/*
 * 㳰ȯƥȤϤ
 */
void printContext(
	INTERRUPT_FRAME *frame)
{
	uint *stack;

	if ((frame->cs == KERNEL_CODE_DES) || (frame->cs == APM_CODE_DES)) {
		printk("error=0x%x eip=0x%x cs=0x%x ds=0x%x esp=0x%x proc=0x%x\n",
			frame->error,
			frame->eip,
			frame->cs,
			frame->ds,
			&frame->user_esp,
			get_current_task());
		stack = &frame->user_esp;
	}
	else{
		printk("error=0x%x eip=0x%x cs=0x%x esp=0x%x,ss=0x%x\n",
			frame->error,
			frame->eip,
			frame->cs,
			frame->user_esp,
			frame->user_ss);
		stack = (uint*) frame->user_esp;
	}
	printk("stack:0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
		stack[0],
		stack[1],
		stack[2],
		stack[3],
		stack[4],
		stack[5],
		stack[6],
		stack[7],
		stack[8],
		stack[9]); 

#ifdef DEBUG
	printSyscallEntry(3);
#endif
}

// 顼
void excDivideError(INTERRUPT_FRAME frame)
{
	exception("Divide error",SIGFPE,&frame);
}


// NMIߡ
void excNonmaskableInterrupt(INTERRUPT_FRAME frame)
{
	char value;


	printk("\nException! Nonmaskable interrupt.\n");

	value=inb(0x61);		/* System control portB. */
	if(value&0x80)
		printk("Memory parity error.\n");
	if(value&0x40)
		printk("IO channel parity error.\n");
	if((value&0xc0)==0)
		printk("Unknown NMI error.\n");

	/* Reset NMI interrupt. */
	outb(0x70,0x8f);
	inb(0x71);
	outb(0x70,0x0f);
	inb(0x71);
}


// Сե
void excOverflow(INTERRUPT_FRAME frame)
{
	exception("Overflow",SIGSEGV,&frame);
}

// BOUNDϰϳ
void excBoundsCheck(INTERRUPT_FRAME frame)
{
	exception("Bounds check",SIGSEGV,&frame);
}

// ̵ڥɡ
void excInvalidOpcode(INTERRUPT_FRAME frame)
{
	exception("Invalid opcode",SIGILL,&frame);
}

// ǥХԲ㳰
void excCoprocessorNotAvailable(INTERRUPT_FRAME frame)
{
	doCoprocessorNotAvailable();
}

// ֥եȡ
void excDoubleFault(INTERRUPT_FRAME frame)
{
	exception("Double fault",SIGSEGV,&frame);
}

// ץåС
void excCopressorOverrun(INTERRUPT_FRAME frame)
{
	exception("Copressor segment overrun",SIGFPE,&frame);
}

// ̵TSS
void excInvalidTss(INTERRUPT_FRAME frame)
{
	exception("Invalid TSS",SIGSEGV,&frame);
}

// Ժߡ
void excSegmentNotPresent(INTERRUPT_FRAME frame)
{
	exception("Segment not present",SIGBUS,&frame);
}

// åȥեȡ
void excStackException(INTERRUPT_FRAME frame)
{
	exception("Stack exception",SIGBUS,&frame);
}

// ݸ
void excGeneralProtection(INTERRUPT_FRAME frame)
{
	exception("General protection",SIGSEGV,&frame);
}

// FPUư顼
void excFpuError(INTERRUPT_FRAME frame)
{
	exception("FPU error",SIGFPE,&frame);
}

// 饤ȥå
void excAlignmentCheck(INTERRUPT_FRAME frame)
{
	exception("Alignment check",SIGBUS,&frame);
}

// ޥå
void excMachineCheck(INTERRUPT_FRAME frame)
{
	/* ̤ */
	tmp_except("Machine check", &frame);
}

// SIMDư㳰
void excSimdFpuError(INTERRUPT_FRAME frame)
{
	exception("SIMD FPU error",SIGFPE,&frame);
}

// ͽ㳰
void excReserve(INTERRUPT_FRAME frame)
{
	tmp_except("15", &frame);
}

/************************************************************************************************
 *
 * ؿ
 *
 ************************************************************************************************/

//================================== PUBLIC =============================================

void init_except()
{
	set_idt( 0, except0,  IDT_TRAP);
	set_idt( 1, except1,  IDT_INTR);
	set_idt( 2, except2,  IDT_TRAP);
	set_idt( 3, except3,  IDT_INTR);
	set_idt( 4, except4,  IDT_TRAP);
	set_idt( 5, except5,  IDT_TRAP);
	set_idt( 6, except6,  IDT_TRAP);
	set_idt( 7, except7,  IDT_INTR);
	set_idt( 8, except8,  IDT_TRAP);
	set_idt( 9, except9,  IDT_TRAP);
	set_idt(10, except10, IDT_TRAP);
	set_idt(11, except11, IDT_TRAP);
	set_idt(12, except12, IDT_TRAP);
	set_idt(13, except13, IDT_TRAP);
	set_idt(14, except14, IDT_INTR);
	set_idt(15, except15, IDT_TRAP);
	set_idt(16, except16, IDT_TRAP);
	set_idt(17, except17, IDT_TRAP);
	set_idt(18, except18, IDT_TRAP);
}
