/*
 * AggregateSchedule.c
 *
 * Copyright 2007, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 *Գס
 * IteratorѥΥץ塼۴ĥꥹ
 *
 *
 * Ǥ¾ϸƤӽФ¦ǤԤȡ
 */


#include <sys/config.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/object.h>
#include <kern/IteratorList.h>
#include <kern/AggregateList.h>

#include <kern/debug.h>


//#define DEBUG_AGGREGATE_SCHEDULE 1
#ifdef DEBUG_AGGREGATE_SCHEDULE
	#define STATIC
	#define INLINE
#else
	#define STATIC	static
	#define INLINE	inline
#endif


//================================== PRIVATE ============================================

/*
 * ³Ǥμ³
 */
STATIC INLINE void addNext(
	List *dst,
	List *src)
{
	src->next = dst->next;
	src->prev = dst;
	src->next->prev = src;
	dst->next = src;
}

/*
 * ³Ǥ³
 */
STATIC INLINE void addPrev(
	List *dst,		// ³ꥹȥСݥ
	List *src)		// ³ꥹȥСݥ
{
	src->next = dst;
	src->prev = dst->prev;
	dst->prev->next = src;
	dst->prev = src;
}

/*
 * ꥹȤ
 */
STATIC INLINE void erase(
	List *src)
{
	src->prev->next = src->next;	// nextǽѤ
	src->next->prev = src->prev;
	src->next =  NULL;
	src->prev = NULL;
}

//================================== PUBLIC =============================================

/*
 * ꥹȥإå³
 *³ꥹȥȥ꡼ϽƤ뤳
 */
STATIC void insertHead(
	AggregateList *this,
	List *m_list)
{
	if (this->head == NULL) {
		m_list->next = m_list->prev = m_list;
		this->head = m_list;
	}
	else {
		addPrev(this->head, m_list);
		this->head = m_list;
	}
}

/*
 * ꥹȥȥ꡼μ³
 */
STATIC void insertNext(
	AggregateList *this,
	List *m_dst,				// ³ȥ꡼
	List *m_src)
{
	/*
	 * ۴ĤϤ줿ȥ꡼μˤ³Ǥ褦ˤ
	 */
	List *prev = m_dst->next->prev;
	
	addNext(prev, m_src);
	m_dst->next = m_src;
}

/*
 * ꥹȥȥ꡼³
 */
STATIC void insertPrev(
	AggregateList *this,
	List *m_dst,				// ³ȥ꡼
	List *m_src)
{
	// ³ꥹȥȥ꡼ϽƤ뤳
	addPrev(m_dst, m_src);
}

/*
 * ꥹȥȥ꡼
 */
STATIC void removeEntry(
	AggregateList *this,
	List *m_list)
{
	// ȥ꡼³Ƥ뤳
	ASSERT(m_list != NULL);

	if (this->head == m_list) {
		if (m_list->next == m_list) {
			this->head = NULL;
		}
		else {
			this->head = m_list->next;
		}
	}
	erase(m_list);
}

/*
 * إåɥȥ꡼Ф
 * return : ȥ꡼ν°֥ or NULL
 */
STATIC OBJECT *getHead(
	AggregateList *this)
{
	if (this->head != NULL) {
		List *headEntry = this->head;
		removeEntry(this, headEntry);

		return headEntry->object;
	}
	else {
		return NULL;
	}
}

/*
 * إåɥȥ꡼򻲾
 * return : ȥ꡼ν°֥ or NULL
 */
STATIC OBJECT *refHead(
	const AggregateList *this)
{
	if (this->head != NULL) {
		return this->head->object;
	}
	else {
		return NULL;
	}
}

/*
 * Υȥ꡼򻲾Ȥ
 * return : ȥ꡼ν°֥ or NULL
 */
STATIC OBJECT *refNext(
	const AggregateList *this,
	const List *i_dst)
{
	ASSERT(i_dst != NULL);

	return i_dst->next->object;
}

/*
 * Υȥ꡼򻲾Ȥ
 * return : ȥ꡼ν°֥ or NULL
 */
STATIC OBJECT *refPrev(
	const AggregateList *this,
	const List *i_dst)
{
	ASSERT(i_dst != NULL);

	return i_dst->prev->object;
}

/*
 * إåλؤȥ꡼򼡤Υȥ꡼˰ܤ
 */
void pointHeadNext(
	AggregateList *this)
{
	ASSERT(this->head != NULL);
	
	this->head = this->head->next;
}

/*
 * ȥ꡼Υȥ꡼˻ؤȥ꡼򡢳ȥ꡼μΥȥ꡼Ѥơ
 * ۴ĥꥹȤϤ
 */
void removeOut(
	const struct AggregateList *this,
	List *m_list)
{
	m_list->prev->next = m_list->next;
	m_list->next->prev = m_list->prev;
}

//---------------------------- ƥ졼 ------------------------------------

STATIC void iterator(
	AggregateList *this,
	IteratorList *m_iterat)
{
	IteratorListConstruct(m_iterat, this);
}

/*
 * إåɥꥹȥȥ꡼򻲾
 * return : ꥹȥȥ꡼ or NULL
 */
STATIC List *refHeadEntry(
	const AggregateList *this)
{
	return this->head;
}

/*
 * Υꥹȥȥ꡼򻲾
 * return : ꥹȥȥ꡼ or NULL
 */
STATIC List *refNextEntry(
	const AggregateList *this,
	const List *i_dst)
{
	ASSERT(i_dst != NULL);

	if (i_dst->next != this->head) {
		return i_dst->next;
	}
	else {
		return NULL;
	}
}

/*
 * ̤б
 */
STATIC void dummyFunc()
{
	printk("This operation not supported!\n");
	ASSERT(0);
}

//---------------------------- 󥹥ȥ饯 ------------------------------------

void AggregateScheduleConstructor(
	AggregateList *this)
{
	// 
	this->head = NULL;
	this->end = NULL;
	this->insertHead = insertHead;
	this->insertEnd = (LIST_FUNC2) dummyFunc;
	this->insertNext = insertNext;
	this->insertPrev = insertPrev;
	this->removeEntry = removeEntry;
	this->getHead = getHead;
	this->getEnd = (LIST_FUNC1) dummyFunc;
//	this->getNext = getNext;
//	this->getPrev = getPrev;
	this->refHead = refHead;
	this->refEnd = (LIST_FUNC6) dummyFunc;
	this->refNext = refNext;
	this->refPrev = refPrev;
	this->pointHeadNext = pointHeadNext;
	this->removeOut = removeOut;
	this->iterator = iterator;
	this->iteratorReverce = (LIST_FUNC9) dummyFunc;
	this->refHeadList = refHeadEntry;
	this->refEndList = (LIST_FUNC10) dummyFunc;
	this->refNextList = refNextEntry;
	this->refPrevList = (LIST_FUNC11) dummyFunc;
}
