/*
 * proc.c
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * ץط
 */


#include <sys/config.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/times.h>
#include <machine/except.h>
#include <machine/segment.h>
#include <machine/lib.h>
#include <machine/tss.h>
#include <machine/fpu.h>
#include <machine/mode_switch.h>
#include <machine/cpuFunc.h>
#include <dev/console/console.h>
#include <net/net.h>
#include <kern/mm.h>
#include <kern/fs.h>
#include <kern/errno.h>
#include <kern/lock.h>
#include <kern/time.h>
#include <kern/timer.h>
#include <kern/elf.h>
#include <kern/signal.h>
#include <kern/term.h>
#include <kern/AggregateList.h>
#include <kern/proc.h>

#include <kern/debug.h>


//#define DEBUG_PROC 1
#ifdef DEBUG_PROC
	#define STATIC
	#define INLINE
#else
	#define STATIC static
	#define INLINE inline
#endif


/******************************************************************************************************
 *
 * CPU
 *
 *¡ե塼뤫κɬץΤ߹Ԥȡ
 *******************************************************************************************************/

//================================== PRIVATE ============================================

/*
 * CPU¤
 */
typedef struct {
	int cpu;					// CPUֹ棰
	uint proc_num;				// ư
	AggregateList aggregate;	// 塼
	int gate;					// ԥå
} TASK_CPU;

static TASK_CPU cputask[MAX_CPU];	// CPU

/*
 * ֥ξʤCPU
 * retuuns : cpu number
 */
STATIC int GetFewTaskCpu()
{
	int cpu = 0;
	int cpuNum = getCpuNum();
	uint num;
	int i;

	num = cputask[0].proc_num;
	for (i = 1; i < cpuNum; ++i) {
		if (cputask[i].proc_num<num) {
			num = cputask[i].proc_num;
			cpu = i;
		}
	}

	return cpu;
}

//================================== PUBLIC =============================================

/*
 *¡CPU˥ꥢꤹɬפ
 * returns : NOERR
 */
int init_cputask(
	int cpu)
{
	cputask[cpu].cpu = cpu;
	cputask[cpu].proc_num = 0;
	cputask[cpu].gate = 0;
	AggregateScheduleConstructor(&cputask[cpu].aggregate);

	return NOERR;
}

/******************************************************************************************************
 *
 * 塼
 *
 *¡ե塼뤫κɬץΤ߹Ԥȡ
 *******************************************************************************************************/

//================================== PRIVATE ============================================

static PROC *idleProc[MAX_CPU];		// ɥץ¹ԥʤ˵ưץ
static PROC *proc1;					// ǽΥץΥץ¤

/*
 *ճߥåϸƤӽФ¦ǹԤ
 * 򥿥塼³
 * return : YES = ³NO = ³Ƥʤ
 */
STATIC int addTask(
	PROC *proc,
	const int cpu,
	const int state)
{
	/* ȥե饰γǧ */
	if ((proc->state & state) != 0) {
		AggregateList *aggregate = &cputask[cpu].aggregate;
		PROC *current = aggregate->refHead(aggregate);

		proc->state = TASK_RUNNING;

		/*
		 * ɥץ³ƤϥɥץϤ
		 */
		if(current == idleProc[cpu]) {
			/*
			 * ɥץ۴ĤϤʥɥץμĤ۴Ĥˤʤ
			 *֡եȤɥ
			 *ȤɥǼʹߤĤ۴ĤˤʤäƤ
			 */
			aggregate->insertNext(aggregate, &current->list, &proc->list);
			aggregate->removeOut(aggregate, &current->list);
		}
		else {
			/*
			 *֡̾ν۴
			 *Ȥ۴ĤϤƤ
			 *Ȥ۴ĤϤƤƼɥץ
			 *嵭ξξ֤ǥȤץ
			 */
			if (proc == current) {
				aggregate->removeEntry(aggregate, &proc->list);
				aggregate->insertHead(aggregate, &proc->list);
			}
			else {
				aggregate->insertNext(aggregate, &current->list, &proc->list);
			}

			// ɥץϥ塼뤫鳰
			if (idleProc[cpu] == aggregate->refNext(aggregate, &proc->list)) {
				aggregate->removeEntry(aggregate, &idleProc[cpu]->list);
			}
		}

		return YES;
	}
	else {
		return NO;
	}
}

/*
 *ճߥåϸƤӽФ¦ǹԤ
 * 򥿥塼뤫ڤΥ
 */
STATIC void delTask(
	const int state)
{
	int cpu = getCpu();

	AggregateList *aggregate = &cputask[cpu].aggregate;
	PROC *current = aggregate->refHead(aggregate);

	/*
	 * ¹ԥʤʤϥɥץ³
	 */
	if(aggregate->refNext(aggregate, &current->list) == current){
		aggregate->insertNext(aggregate, &current->list, &idleProc[cpu]->list);
	}
	aggregate->removeOut(aggregate, &current->list);

	current->state = state;
}

//================================== PUBLIC =============================================

/*
 * Get current task
 * return : Current process
 */
PROC *get_current_task()
{
	return cputask[getCpu()].aggregate.refHead(&cputask[getCpu()].aggregate);
}

/*
 * å
 * return : Page directory
 */
uint *switch_task(
	uint esp,					// γ߻Υ֥åȥåesp
	volatile uint beforeEsp)	// γ߻espͤ¸륹å
{
	int cpu = getCpu();
	uint volatile *p_beforeEsp = &beforeEsp;
	uint volatile *p_esp = &esp;
	AggregateList *aggregate = &cputask[cpu].aggregate;
	PROC *current;

	current = aggregate->refHead(aggregate);
	*p_beforeEsp = current->esp;
	current->esp = esp;

	// Υ˥å
	enter_spinlock(&cputask[cpu].gate);
	{
		aggregate->pointHeadNext(aggregate);
		current = aggregate->refHead(aggregate);
	}
	exit_spinlock(&cputask[cpu].gate);

	*p_esp = (uint) &current->esp;

	return getPageDir();
}

/*
 * ¹ԥ塼³롣
 *¾ץƤФ롣
 *Գ
 */
void addToSchedule(
	PROC *proc, 
	const int i_state)		// "TASK_RUNNING"ʤ"o_isDel""PROC_SCHEDULE_NOT_DEL"
{
	int cpu = proc->cpu;
	int eflags;

	eflags = enterCli();
	enter_spinlock(&cputask[cpu].gate);
	{
		addTask(proc, cpu, i_state);
	}
	exit_spinlock(&cputask[cpu].gate);
	exitCli(eflags);
}

/*
 * 򤹤¹ԤǤ褦ƱCPU˰ܤ
 *¾ץƤФ롣夹˥å롣
 *Գ
 */
void addScheduleSoon(
	PROC *proc,
	const int i_state)
{
	int cpu = getCpu();
	int eflags;

	eflags = enterCli();
	enter_spinlock(&cputask[cpu].gate);
	{
		if (addTask(proc, cpu, i_state) == YES) {
			--cputask[proc->cpu].proc_num;
			++cputask[cpu].proc_num;
			proc->cpu = cpu;
		}
	}
	exit_spinlock(&cputask[cpu].gate);
	exitCli(eflags);
}

/*
 * ¹ԥ塼뤫ڤΥ
 *ռץƤ֡
 *ԻڤΥ奿åޤǳߤʤ褦˳߶ػߤˤ롣
 */
void delFromSchedule(
	const int i_state)
{
	int cpu = getCpu();

	cli();
	enter_spinlock(&cputask[cpu].gate);
	{
		delTask(i_state);
	}
	exit_spinlock(&cputask[cpu].gate);
}

/*
 * ưե饰ꤷƻư
 *¾ץƤФ롣
 *Գ
 */
void addScheduleSetFlag(
	PROC *proc,
	const int i_state,
	int *m_isAdd)			// ȥߥåYESꤹ
{
	int cpu = proc->cpu;
	int eflags;

	eflags = enterCli();
	enter_spinlock(&cputask[cpu].gate);
	{
		addTask(proc, cpu, i_state);
		*m_isAdd = YES;
	}
	exit_spinlock(&cputask[cpu].gate);
	exitCli(eflags);
}

/*
 * ưե饰ˤ꥿¹ԥ塼뤫ڤΥ
 *ռץƤ֡
 *ԻڤΥ奿åޤǳߤʤ褦˳߶ػߤˤ롣
 */
void delFromScheduleByFlag(
	const int i_state,
	int volatile *isAdd)	// ˥³줿 YES of NO
{
	int cpu = getCpu();
	int isDel;
	int eflags;
	
	eflags = enterCli();
	enter_spinlock(&cputask[cpu].gate);
	{
		if (*isAdd == NO) {
			delTask(i_state);
			isDel = YES;
		}
		else {
			isDel = NO;
		}
	}
	exit_spinlock(&cputask[cpu].gate);

	if (isDel == NO) {
		exitCli(eflags);
	}
}

/******************************************************************************************************
 *
 * ץꥹ
 *
 *******************************************************************************************************/

//================================== PRIVATE ============================================

enum{
	PROC_HASH_NUM=29,	/* ץϥåꥹȺ */
};

static PROC_LIST procHash[PROC_HASH_NUM];		/* ץϥåꥹȡ */

/*
 * ץꥹͤץɥ쥹롣
 * return : process address
 */
STATIC INLINE PROC *procLinkToProc(
	PROC_LIST *p)		// process link address
{
	return (PROC*)((uint)p-(uint)&((PROC*)0)->proc_next);
}

/*
 * ϥåͤ롣
 * return : hash value
 */
STATIC INLINE int getProcHash(
	pid_t pid)
{
	return pid % PROC_HASH_NUM;
}

/*
 * Add to process list.
 */
STATIC void addProcList(
	PROC *proc)
{
	PROC_LIST *hash = &procHash[getProcHash(proc->pid)];
	PROC_LIST *p = (PROC_LIST*)&proc->proc_next;

	enter_spinlock(&hash->lockGate);
	{
		p->prev = hash;
		p->next = hash->next;
		hash->next->prev = p;
		hash->next = p;
	}
	exit_spinlock(&hash->lockGate);
}

/*
 * Delete from process list.
 */
STATIC void delProcList(
	PROC *proc)
{
	PROC_LIST *hash=&procHash[getProcHash(proc->pid)];
	PROC_LIST *p=(PROC_LIST*)&proc->proc_next;

	enter_spinlock(&hash->lockGate);
	{
		p->next->prev=p->prev;
		p->prev->next=p->next;
	}
	exit_spinlock(&hash->lockGate);
}

/*
 * Init process list;
 */
STATIC void initProclist()
{
	int i;

	/* ץϥåꥹȽ */
	for (i = 0; i < PROC_HASH_NUM; ++i) {
		procHash[i].next=procHash[i].prev = &procHash[i];
		procHash[i].lockGate = 0;
	}
}

//================================== PUBLIC =============================================

/*
 * Search process from process list.
 * return : process address or failed=NULL
 */
PROC *searchProcList(
	pid_t pid)
{
	PROC_LIST *hash=&procHash[getProcHash(pid)];
	PROC_LIST *p;

	for (p = hash->next; p != hash; p = p->next) {
		if (procLinkToProc(p)->pid == pid) {
			return procLinkToProc(p);
		}
	}

	return NULL;
}

/******************************************************************************************************
 *
 * ץID롼ID
 *
 *******************************************************************************************************/

//================================== PRIVATE ============================================

/*
 * ץ롼פθ
 * return : process or NULL
 */
STATIC PROC *searchPg(
	PROC *proc,
	pid_t pgid)
{
	PROC *rest;

	do {
		if (proc->state != TASK_EXIT) {
			if (proc->pgid == pgid) {
				return proc;
			}
			if (proc->child != NULL) {
				if ((rest = searchPg(proc->child, pgid)) != NULL) {
					return rest;
				}
			}
		}
	} while((proc = proc->brother) != NULL);

	return NULL;
}

/*
 * ֲλҥץ饷ʥ
 * return : prosecc or Failed=NULL
 */
STATIC void sendToProc(
	PROC *proc,		// search start process or NULL=proc1
	int signal)
{
	/* μν֤Υץ */
	if (proc->child != NULL) {
		sendToProc(proc->child, signal);
	}
	if (proc->brother != NULL) {
		sendToProc(proc->brother, signal);
	}

	if (proc != get_current_task()) {
		sendSignal(proc,signal);
	}
}

/*
 *¡ĹϢ³Ư硢Сեˤʤǽꡣ
 * Get process ID.
 * return : process ID.
 */
STATIC INLINE int getPid(PROC *proc)
{
	static int pid = 1;

	return pid++;
}

//================================== PUBLIC =============================================

/*
 * Ϳ줿ץ롼פΥץ롣
 * return : prosecc or Failed=NULL
 */
PROC *getNextProc(
	PROC *proc,		// search start process or NULL=proc1
	pid_t pgid)
{
	/* proc1õ */
	if (proc == NULL) {
		proc = proc1;
	}

	/* μν֤Υץõ */
	if (proc->child != NULL) {
		return searchPg(proc->child,pgid);
	}
	else {
		do {
			if (proc->brother != NULL) {
				return searchPg(proc->brother,pgid);
			}
			if ((proc = proc->parent) == NULL) {
				break;
			}
		} while (proc->pgid == pgid);
	}

	return NULL;
}

/*
 * ץʳץ˥ʥ
 */
void sendSignalAllProc(
	int signal)
{
	if (proc1->child != NULL) {
		sendToProc(proc1->child, signal);
	}
}

/*
 * ѡ桼
 * return : error number
 */
int suser(PROC *proc)
{
	return (proc->uid == 0) ? NOERR : -EPERM;
}

/***************************************************************************************************
 *
 * ץ
 *
 ***************************************************************************************************/

//================================== PRIVATE ============================================

// CPU֤ƥץγΰ­
STATIC void addCpuTimeToParent()
{
	PROC *proc = get_current_task();
	PROC *parent = proc->parent;

	parent->ctime[0] += calcMiliSecondFromClock(proc->sysClocks) / TASK_TIME;
	parent->ctime[1] += calcMiliSecondFromClock(proc->userClocks) / TASK_TIME;
}

//================================== PUBLIC =============================================

/*
 *ճߥͥȥȲûޤϸ˸ƤӽФ
 * ץַ¬rdtsc͡ˤꤹ
 */
void setProcCurrentTime()
{
	if (isNest() == NO) {
		get_current_task()->beforeClocks = rdtsc();
	}
}

/*
 *ճߥͥȥȲû˸ƤӽФ
 * ץ桼֤¬
 */
void calcProcUserTime()
{
	if (isNest() == NO) {
		int64 userClocks = rdtsc() - get_current_task()->beforeClocks;

		/*
		 * ¹CPUѤsysClocksޥʥˤʤ뤳Ȥ⤢ꤨ
		 */
		get_current_task()->userClocks += userClocks;
		doUserItimer(userClocks);
	}
}

/*
 *ճߥͥȥȸ˸ƤӽФ
 * ץƥ֤¬
 */
void calcProcSysTime()
{
	if (isSingleNest() == YES) {
		int64 sysClocks = rdtsc() - get_current_task()->beforeClocks;
		
		/*
		 * ¹CPUѤsysClocksޥʥˤʤ뤳Ȥ⤢ꤨ
		 */
		get_current_task()->sysClocks += sysClocks;
		doSystemItimer(sysClocks);
	}
}

/******************************************************************************************************
 *
 * ץ
 *
 *******************************************************************************************************/

static int procLockGate;		/* ¾Υץ¤λȻΥåȡ */


//================================== PRIVATE ============================================

/*
 * λҥץξ֤𤹤롣
 * ӥåȥޥåװ֤ΰ̣"sys/wait.h"򻲾ȤΤ
 * return : 0 or not exit or stop=-1
 */
STATIC void setState(
	PROC *proc,
	int *stat_loc)		// state address
{
	union{
		int i;
		char c[4];
	}stat;
	
	if (stat_loc == NULL)
		return;

	stat.i = 0;
	stat.c[0] = proc->signum;
	stat.c[1] = proc->exit_state;
	if (proc->state == TASK_SIGNAL_STOP) {
		stat.c[0] = 0x7F;
		stat.c[1] = proc->signum;
	}
	*stat_loc = stat.i;
}

/*
 * exitҥץ¤Τ롣
 * parameters : current process
 * return : 󥯤κǸλҥץ
 */
STATIC PROC *releaseChild(
	PROC *proc)
{
	PROC **p,*q;

	for (p = &proc->child; *p != NULL;) {
		if ((*p)->state == TASK_EXIT) {
			q = *p;
			*p = (*p)->brother;
			kfree(q);
		}
		else {
			p = &(*p)->brother;
		}
	}

	return *p;
}

/*
 * ץ¹Բǽˤ
 */
STATIC void addNewProc(
	PROC *proc)
{
	proc->cpu = GetFewTaskCpu();
	proc->state = TASK_WAIT;
	addToSchedule(proc, TASK_WAIT);
	++cputask[proc->cpu].proc_num;
}


/*
 * ƥץ¹ԡ
 * 1 exitҥץγ
 * 2 Ĥäҥץpid0ץϤ
 * parameters : current process
 */
STATIC void releaseChildProc(PROC *proc)
{
	PROC *p,**q;

	for (p = proc->child; p != NULL;p = p->brother){
		p->parent = proc1;
	}

	enter_spinlock(&proc->gate);
	{
		p = releaseChild(proc);
	}
	exit_spinlock(&proc->gate);

	enter_spinlock(&proc1->gate);
	{
		for (q = &proc1->child; *q != NULL; q = &(*q)->brother);
		*q = p;
	}
	exit_spinlock(&proc1->gate);
}

/*
 * 1 exitҥץ롣
 * 2 λҥץξ֤𤹤롣
 * parameters : process structure,search pid,search pgid,task state,state address
 * return : pid or 0 or error number
 */
STATIC int searchExitState(PROC *proc, pid_t pid, pid_t pgid, int state, int *stat_loc)
{
	int rest = -ECHILD;
	PROC *p;

	if (stat_loc != NULL){
		*stat_loc = 0;
	}

	if (pid != 0) {
		for (p = proc->child; p != NULL; p = p->brother) {
			if (p->pid == pid) {
				rest = 0;
				if (p->state & state) {
					setState(p,stat_loc);
					rest = pid;
				}
				break;
			}
		}
	}
	else if (pgid != 0) {
		for (p = proc->child; p != NULL; p = p->brother) {
			if (p->pgid == pgid) {
				rest = 0;
				if (p->state & state) {
					setState(p,stat_loc);
					rest = p->pid;
					break;
				}
			}
		}
	}
	else if (proc->child != NULL) {
		rest = 0;
		for (p = proc->child; p != NULL; p = p->brother) {
			if (p->state & state) {
				setState(p,stat_loc);
				rest = p->pid;
				break;
			}
		}
	}

	/* exitҥץγ */
	enter_spinlock(&proc->gate);
	{
		releaseChild(proc);
	}
	exit_spinlock(&proc->gate);

	return rest;
}

/*
 * ץե
 */
STATIC int fork(
	int isKernelThread)		// isKernelThreadˤϥͥ륹åɤȤ"0"ʳ
{
	void *stack;
	uint esp;
	PROC *parent = get_current_task();
	PROC *child;
	int rest;

	if (MAX_CHILD <= parent->count_child) {
		return -EAGAIN;
	}

	/* PROC¤ΤƤ */
	child = kmalloc(sizeof(PROC));
	if (child == NULL) {
		return -ENOMEM;
	}

	/*  */
	memset(&child->PROC_ZERO_TOP, 0, (uint) &child->PROC_ZERO_END - (uint) &child->PROC_ZERO_TOP);
	listConstructor(&child->list, child);
	listConstructor(&child->waitList, child);
	child->parent = parent;
	enter_spinlock(&parent->gate);
	{
		child->brother = parent->child;
		parent->child = child;
	}
	exit_spinlock(&parent->gate);
	child->nestCount = parent->nestCount;

	/* ID. */
	child->pid = getPid(child);
	child->pgid = parent->pgid;
	child->uid = 0;
	child->gid = parent->gid;
	child->sid = parent->sid;

	/* å󥢥ɥ쥹(ü¤Υɥ쥹)򥳥ԡ */
	child->ctlterm = parent->ctlterm;
	
	child->gate = 0;

	/* եط¤Τγơ */
	child->file_struct = copyFileStruct(parent->file_struct);
	if ((child->file_struct) == NULL){
		goto ERR;
	}

	/* Page tableƤ */
	rest = forkPage(parent->mm_struct, isKernelThread, &child->mm_struct, &stack);
	if (rest < 0){
		goto ERR;
	}

	/* ʥꤹ롣 */
	child->signal_struct = setSignal(parent->signal_struct);
	if (child->signal_struct == NULL){
		goto ERR;
	}

	/* ޡν */
	initProcTimer(child);

	/* ҥץΥͥ륹å˥ƥȤꤹ */
	esp = setChildContext(stack);
	if (esp == 0){
		return 0;		// ҥץ
	}

	child->esp = esp;

	/* ץꥹȤ˲ä롣 */
	addProcList(child);

	// ץ塼˲ä
	addNewProc(child);

	++parent->count_child;

	return child->pid;
ERR:
	kfree(child);

	return -ENOMEM;
}

/*
 * ץλ
 */
void exit(
	int state,		// ƥץϤ
	int signum)
{
	void *mm_struct;
	PROC *proc=get_current_task();
	PROC *parent;

	/* üץγ */
	releaseTerm(proc);

	/* ץꥹȤ롣 */
	delProcList(proc);

	/* ҥץγ */
	releaseChildProc(proc);

	parent = proc->parent;

	--parent->count_child;

	/* եǥץγ */
	releaseFileStruct(proc->file_struct);

	/* ޡγ */
	releaseProcTimer();

	/* ʥγ */
	releaseSignal(proc->signal_struct);

	mm_struct = proc->mm_struct;
	proc->exit_state = state & 0xff;
	proc->signum = signum & 0xff;

	// FPU֥ɥ쥹򥯥ꥢ
	clearFpuSaveAddr();

	/* CPUƥץ­ */
	addCpuTimeToParent();

	/* 塼뤫waitƤƤƤӽФ */
	enter_spinlock(&procLockGate);
	{
		{
//			static PROC tmp_proc[MAX_CPU];
			int cpu = getCpu();

			delFromSchedule(TASK_EXIT);
//			tmp_proc[cpu].next=proc->next;
//			tmp_proc[cpu].prev=proc->prev;
//			cputask[cpu].current_task=&tmp_proc[cpu];
			--cputask[cpu].proc_num;
		}

		sendSignal(parent,SIGCHLD);
//		if ((isForeground(proc) == YES) && (parent->state & TASK_CHILD_WAIT))
		if ((parent->state & TASK_CHILD_WAIT) == TASK_CHILD_WAIT){
			addToSchedule(parent, TASK_SIGNAL_WAIT | TASK_SIGNAL_STOP);
		}
	}
	exit_spinlock(&procLockGate);

	/* Υ˥åơڡǥ쥯ȥ롣 */
	returnExit(mm_struct);
}

//================================== PUBLIC =============================================

/*
 * Set idle process table
 * returns : Error number
 */
int createIdleProc(
	int cpu)
{
	int error;

	/* Set process table */
	idleProc[cpu] = kmalloc(sizeof(PROC));
	if (idleProc[cpu] == NULL) {
		return -ENOMEM;
	}
	memset(idleProc[cpu], 0, sizeof(PROC));
	listConstructor(&idleProc[cpu]->list, idleProc[cpu]);
	listConstructor(&idleProc[cpu]->waitList, idleProc[cpu]);
	idleProc[cpu]->cpu = cpu;
	idleProc[cpu]->state = TASK_RUNNING;

	/* init paging */
	idleProc[cpu]->mm_struct = initPaging();
	if (idleProc[cpu]->mm_struct == NULL) {
		return -ENOMEM;
	}

	/*
	 * FPU֤ν
	 * ʹߤλҥץˤ⥳ԡΤǡ
	 * ǽ˥ץFPUԲ㳰򵯤FPUȥ
	 * Ƥ۾򤪤ʤ
	 */
	fpuSave(KERNEL_SAVE_FPU);

	/* init file descriptor. */
	error = init_file_struct(idleProc[cpu]);
	if (error != NOERR) {
		return error;
	}

	/* init signal. */
	idleProc[cpu]->signal_struct = setFirstSignal();
	if (idleProc[cpu]->signal_struct == NULL) {
		return -ENOMEM;
	}

	/* add to schedule queue */
	cputask[cpu].aggregate.insertHead(&cputask[cpu].aggregate, &idleProc[cpu]->list);

	/* ץꥹȤν */
	initProclist();

	return NOERR;
}

/*
 * ǽΥץꡣ
 * return : 0 or error number
 */
int initProc1()
{
	proc1 = get_current_task();
	proc1->parent = NULL;

	return initProc1Fs();
}

/*
 * proc1wait
 */
void waitProc1()
{
	for(;;)
	{
		/* Ԥ */
		delFromSchedule(TASK_SIGNAL_WAIT);
		wait_task();

		enter_spinlock(&proc1->gate);
		{
			releaseChild(proc1);
		}
		exit_spinlock(&proc1->gate);
	}
	idle();
}

/*
 * ͥͥȥȤ䤹
 */
void incrementNest()
{
	get_current_task()->nestCount += 1;
}

/*
 * ͥͥȥȤ򸺤餹
 */
void decrementNest()
{
	get_current_task()->nestCount -= 1;

	ASSERT(0 <= get_current_task()->nestCount);
}

/*
 * ͥǥͥȤƤ뤫
 * return : YES or NO
 */
int isNest()
{
	return (0 < get_current_task()->nestCount) ? YES : NO;
}

/*
 * ͥǥͥȲǤ뤫
 * return : YES or NO
 */
int isSingleNest()
{
	return (1 == get_current_task()->nestCount) ? YES : NO;
}

/*
 * ͥ륹åɤfork
 */
int forkKernelThread()
{
	return fork(1);
}

/*
 * ͥ륹åɤ
 * return : error number or 0
 */
int kthread_create(void (*func)(void *), void *arg, struct proc **newpp, const char *fmt, ...)
{
	int pid = forkKernelThread();

	switch(pid){
	case 0:
		func(arg);
		exit(0, 0);
	case -1:
		return -ENOBUFS;
	default:
		*newpp = searchProcList(pid);
		return 0;
	}
}

/*
 * ͥ륹åɤλ
 */
void kthread_exit(int ecode)
{
	exit(ecode, 0);
}

/******************************************************************************************************
 *
 * ƥॳ
 *
 *******************************************************************************************************/

/*
 * return : ƥץʤҥץID,ҥץʤ0,error=error number
 * todoꥹ
 *  ǥ쥯ȥꥹȥ꡼򥳥ԡ롣
 */
int sys_fork()
{
	return fork(0);
}

int sys_exit(int state)
{
	exit(state,0);

	return 0;
}

int sys_wait(int pid, int *stat_loc, int options)
{
	int rest;
	int state;
	PROC *proc = get_current_task();

	/* Ĵ٤ץ֤ꡣ */
	state = TASK_TRACE | TASK_EXIT;
	if (options & WUNTRACED) {
		state = TASK_SIGNAL_STOP | TASK_EXIT;
	}

	/* ǽλҥץ */
	if (0 < pid) {
		rest = searchExitState(proc, pid, 0, state, stat_loc);
	}
	else if (pid == 0) {
		rest = searchExitState(proc, 0, proc->pgid, state, stat_loc);
	}
	else if (pid == -1) {
		rest = searchExitState(proc, 0, 0, state, stat_loc);
	}
	else {
		rest = searchExitState(proc, 0, -pid, state, stat_loc);
	}

	if (0 < rest) {
		return rest;
	}

	if ((rest < 0) && (options & WNOHANG)) {
		return rest;
	}

	/* Ԥ */
	delFromSchedule(TASK_SIGNAL_WAIT|TASK_CHILD_WAIT);
	wait_task();

	/* ҥץ */
	if (0 < pid) {
		rest = searchExitState(proc, pid, 0, state, stat_loc);
	}
	else if (pid == 0) {
		rest = searchExitState(proc, 0, proc->pgid, state, stat_loc);
	}
	else if (pid == -1) {
		rest = searchExitState(proc, 0, 0, state, stat_loc);
	}
	else {
		rest = searchExitState(proc, 0, -pid, state, stat_loc);
	}

	return rest;
}

int sys_getpid()
{
	return get_current_task()->pid;
}

int sys_getgid()
{
	return get_current_task()->gid;
}

int sys_getpgid(pid_t pid)
{
	if (pid == 0) {
		return get_current_task()->pgid;
	}
	else {
		PROC *proc = searchProcList(pid);
		
		if (proc == NULL) {
			return -ESRCH;
		}
		if (proc->sid != get_current_task()->sid) {
			return -EPERM;
		}
		return proc->pgid;
	}
}

int sys_getppid()
{
	return get_current_task()->parent->pid;
}

int sys_getuid()
{
	return get_current_task()->uid;
}

int sys_getsid(pid_t pid)
{
	if (pid == 0) {
		return get_current_task()->sid;
	}
	else {
		PROC *proc = searchProcList(pid);
		
		if (proc == NULL) {
			return -ESRCH;
		}
		if (proc->sid != get_current_task()->sid) {
			return -EPERM;
		}
		return proc->sid;
	}
}

int sys_setgid(gid_t gid)
{
	PROC *proc = get_current_task();

	if (proc->uid != 0) {
		return -EPERM;
	}
	proc->gid = gid;
	return 0;
}

int sys_setpgid(pid_t pid, pid_t pgid)
{
	PROC *proc,*pgproc;
	
	if (pid == 0) {
		proc = get_current_task();
	}
	else {
		proc = searchProcList(pid);
		if (proc == NULL) {
			return -ESRCH;
		}
		if ((proc != get_current_task()) && (proc->parent != get_current_task())) {
			return -ESRCH;
		}
	}

	if (pgid < 0) {
		return -EINVAL;
	}
	if (pgid == 0) {
		pgid = proc->pid;
	}
	
	/* Session leader */
	if (proc->sid == proc->pid) {
		return -EPERM;
	}
	if (proc->sid != get_current_task()->sid) {
		return -EPERM;
	}
	
	/* ץۤʤ륻åΥץ롼פ˰ư褦Ȥ */
	pgproc = searchProcList(pgid);
	if (pgproc == NULL) {
		return -EPERM;
	}
	if (pgproc->sid != proc->sid) {
		return -EPERM;
	}
	proc->pgid = pgid;

	return 0;
}

int sys_setuid(uid_t uid)
{
	PROC *proc = get_current_task();

	if (proc->uid != 0) {
		return -EPERM;
	}
	proc->uid = uid;

	return 0;
}

int sys_setsid()
{
	PROC *proc = get_current_task();
	
	if (proc->pgid != proc->pid) {
		proc->sid = proc->pid;
		proc->pgid = proc->pid;
		proc->ctlterm = NULL;
		
		return proc->sid;
	}
	else {
		return -EPERM;
	}
}

int sys_times(struct tms *tms)
{
	PROC *proc;

	if (checkMem(tms, sizeof(struct tms)) == -1) {
		return -EFAULT;
	}

	proc=get_current_task();
	tms->tms_stime = calcMiliSecondFromClock(proc->sysClocks) / TASK_TIME;
	tms->tms_utime = calcMiliSecondFromClock(proc->userClocks) / TASK_TIME;
	tms->tms_cstime = proc->ctime[0];
	tms->tms_cutime = proc->ctime[1];

	return 0;
}

/*************************************************************************************************/
void test_proc(
	int cpu)
{
	enum {
		MAX = 20,
	};
	char str[MAX];
	int len = 0;
	AggregateList *aggregate = &cputask[cpu].aggregate;
	PROC *proc;
	PROC *nextEnd;
	PROC *end;

	memset(str, 0, MAX);
	proc = aggregate->refHead(aggregate);
	nextEnd = aggregate->refHead(aggregate);
	do {
		PROC *prev;

		len += sprintk(&str[len], "%d ", proc->pid);
		if (MAX <= len) {
			break;
		}

		end = nextEnd;
		prev = proc;
		proc = aggregate->refNext(aggregate, &proc->list);
		if (aggregate->refPrev(aggregate, &proc->list) != prev) {
			nextEnd = proc;
		}
	} while (proc != end);
	
	str[MAX - 1] = '\0';
	printDebug(80 * 6 + 60, "%s", str);
}
/**************************************************************************************************/
