#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "AggregateList.h"
#include "IteratorList.h"

#include "test.h"


/******************************* eXgp` ******************************************/

/******************************* eXgR[h ******************************************/

/******************************* _v ******************************************/

typedef struct {
	int number;
	List list;
} ENTRY;

static void printEntry(
	const List *i_list)
{
	printf("%02d-0x%06x-0x%06x-0x%06x\n",
		((ENTRY*) i_list->object)->number,
		i_list,
		i_list->prev,
		i_list->next);
}

static void dump(
	const AggregateList *i_aggregate)
{
	List *entry;
	
	if (i_aggregate->head != NULL) {
		printf("head = %d\n", ((ENTRY*) i_aggregate->head->object)->number);
	}
	else {
		printf("head = NULL\n");
	}
	
	// PEoXgp_v
	for (entry = i_aggregate->head; entry != NULL; entry = entry->next) {
		printEntry(entry);
	}

	// zXgp_v
/*	entry = i_aggregate->head;
	if (entry != NULL) {
		List *end;
		List *nextEnd = i_aggregate->head;
		do {
			List *prev;

			printEntry(entry);

			end = nextEnd;
			prev = entry;
			entry = entry->next;
			if (entry->prev != prev) {
				nextEnd = entry;
			}
		} while (entry != end);
	}
*/
	if (i_aggregate->end != NULL) {
		printf("end = %d\n", ((ENTRY*) i_aggregate->end->object)->number);
	}
	else {
		printf("end = NULL\n");
	}

}

/******************************************************************************************/

enum {
	MAX = 10,
};

static ENTRY entry[MAX];

static void initTest(
	AggregateList *m_aggregate)
{
	int i;
	
	for (i = 0; i < MAX; ++i) {
		entry[i].number = i;
		listConstructor(&entry[i].list, &entry[i]);
	}
}

int main(int argc, char *argv[])
{
	AggregateList aggregate;

	AggregateSingleListConstructor(&aggregate);
//	AggregateScheduleConstructor(&aggregate);
	initTest(&aggregate);

	for (;;) {
		int func;
		int num1;
		int num2;

		do {
			scanf("%d %d %d", &func, &num1, &num2);
			if (MAX <= num1) {
				printf("number1:%d is over from Max\n", num1);
			}
			if (MAX <= num2) {
				printf("number2:%d is over from Max\n", num2);
			}
		} while ((MAX <= num1) || (MAX <= num2));

		switch (func) {
		case 10:
			aggregate.insertHead(&aggregate, &entry[num1].list);
			break;
		case 11:
			aggregate.insertEnd(&aggregate, &entry[num1].list);
			break;
		case 12:
			aggregate.insertNext(&aggregate, &entry[num1].list, &entry[num2].list);
			break;
		case 13:
			aggregate.insertPrev(&aggregate, &entry[num1].list, &entry[num2].list);
			break;
		case 14:
			aggregate.removeEntry(&aggregate, &entry[num1].list);
			break;
		case 15:
			aggregate.pointHeadNext(&aggregate);
			break;
		case 16:
			aggregate.removeOut(&aggregate, &entry[num1].list);
			break;
		case 20: {
			ENTRY *ent = aggregate.getHead(&aggregate);
			if (ent != NULL) {
				printf("getHead = %d\n", ent->number);
			}
			else {
				printf("getHead = NULL\n");
			}
			break;
		}
		case 21: {
			ENTRY *ent = aggregate.getEnd(&aggregate);
			if (ent != NULL) {
				printf("getEnd = %d\n", ent->number);
			}
			else {
				printf("getEnd = NULL\n");
			}
			break;
		}
		case 30: {
			ENTRY *ent = aggregate.refHead(&aggregate);
			if (ent != NULL) {
				printf("refHead = %d\n", ent->number);
			}
			else {
				printf("refHead = NULL\n");
			}
			break;
		}
		case 31: {
			ENTRY *ent = aggregate.refEnd(&aggregate);
			if (ent != NULL) {
				printf("refEnd = %d\n", ent->number);
			}
			else {
				printf("refEnd = NULL\n");
			}
			break;
		}
		case 32: {
			ENTRY *ent = aggregate.refNext(&aggregate, &entry[num1].list);
			if (ent != NULL) {
				printf("refNext = %d\n", ent->number);
			}
			else {
				printf("refNext = NULL\n");
			}
			break;
		}
		case 33: {
			ENTRY *ent = aggregate.refPrev(&aggregate, &entry[num1].list);
			if (ent != NULL) {
				printf("refPrev = %d\n", ent->number);
			}
			else {
				printf("refPrev = NULL\n");
			}
			break;
		}
		case 40: {
			IteratorList *iterator = aggregate.iterator(&aggregate);
			while (iterator->hasNext(iterator) == BOOL_TRUE) {
				ENTRY *ent = iterator->next(iterator);
				printf("%d ", ent->number);
			}
			printf("\n");
			break;
		}
		default:
			printf("Unknown func number : %d!\n", func);
			break;
		}

		dump(&aggregate);
	}

	return 0;
}
