#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "list.h"


enum {
	PAGE_SIZE = 8,
	NOERR = 0,
	EIO,

	NO = 0,
	YES,

	FALSE	= 0,
	TRUE,
};
#define STATIC	static
#define INLINE
#define kmalloc malloc
#define kfree free
#define printk printf
#define ROUNDUP(a, b) (((uint)(a) + (b) - 1) / (b) * (b))
#define ROUNDDOWN(a, b) ((uint)(a) / (b) * (b))
#define ROUNDUP_DIV(a, b) (((uint)(a) + (b) - 1) / (b))
#define ASSERT(bool) if ((bool) == 0){							\
	printf("Assertion! : %s line=%d \n",__FILE__, __LINE__); 	\
}
#define WARNING(bool) if ((bool) == 0){							\
	printf("Warning! : %s line=%d \n",__FILE__, __LINE__);		\
}
#define OFFSETOF(type, member) ((size_t) &((type*)0)->member)
#define ENOMEM 5

#define dev_t struct device*
#define memcpy(a,b,c)
#define wait_task()
typedef unsigned char uchar;
typedef unsigned char u_int8_t;
typedef unsigned int uint;
typedef int PROC;
enum uio_rw { 
	UIO_READ, 
	UIO_WRITE 
};
struct iovec {
	char	*iov_base;	/* Base address. */
	size_t	 iov_len;	/* Length. */
};
struct uio {
	struct iovec	*uio_iov;
	int				uio_iovcnt;
	off_t			uio_offset;
	int				uio_resid;
	enum			uio_rw uio_rw;
};
struct device;
typedef int (*d_read) (dev_t dev, struct uio *uio, int ioflag);
struct disk {
	uint sectors;
	uint beginSector;
};
struct cdevsw {
	d_read		d_read;
	d_read		d_write;
} si_devsw;
struct device {
	struct disk disk;
	struct cdevsw *si_devsw;
} device;

static void enter_spinlock(int *dummy) { return; }
static void exit_spinlock(int *dummy) { return; }

/******************************* eXgR[h ******************************************/

/*
 * ubNLbV
 */
typedef struct BLOCK_CACHE {
	// lock : cacheLock
	int 				din;			// foCXinode.
	uint				sector;			// foCXZN^[ԍiCACHE_SECTORSEj
	char				*cache;			// LbV
	OBJ_LIST			seqList;		// foCXinodeԍEZN^[ԍLbVzXg
	OBJ_LIST			updateList;		// ŐVLbVzXg
	u_int8_t			bitmap;			// LbVrbg}bvAPrbgPZN^[
	u_int8_t			dirtyBitmap;	// XVrbg}bvAPrbgPZN^[

	// lock : writeLock
	OBJ_LIST			writeList;		// ߂҂zXg
	OBJ_LIST			writeSeqList;	// foCXinodeԍEZN^[ԍ߂҂zXg

	// syncronize : writeBusy
	u_int8_t			writeBitmap;	// ߂rbg}bvAPrbgPZN^[
	int 				writeBusy;		// ߂tO
	struct BLOCK_CACHE	*cacheNext;		// ߂҂XgN
} BLOCK_CACHE;

enum {
	CACHE_NUM_MAX = 6,					// IOLbVubN
	CACHE_SIZE = PAGE_SIZE,				// IOLbVubNTCY
	CACHE_SECTOR_SIZE = 1,				// ZN^[TCY
	CACHE_SECTORS = CACHE_SIZE / CACHE_SECTOR_SIZE,
	
	IOVEC_MAX = 4,	// iovec
};

// ubNLbVCX^X
static BLOCK_CACHE blkCacheInstance[CACHE_NUM_MAX];

static int cacheLock;			// ubNLbV샍bN
static int writeLock;			// xݑ҂bN

static OBJ_LIST *seqHead;		// ubNLbVXgwb_
static OBJ_LIST *updateHead;	// ubNLbVXVXgwb_
static OBJ_LIST *writeHead;		// ubNLbVxݑ҂Xgwb_
static OBJ_LIST *writeSeqHead;	// ubNLbVxݑ҂foCXinodeԍEZN^[ԍXgwb_

//============================ _~[ ==================================================

static void printDump(BLOCK_CACHE *blkCache)
{
	printf("0x%06x-%03d-%02x-%02x-%02x-%d ", 
		blkCache,
		blkCache->sector, 
		blkCache->bitmap,
		blkCache->dirtyBitmap,
		blkCache->writeBitmap,
		blkCache->writeBusy);
}

static void listDump(
	OBJ_LIST *listHead,
	const uint offset)
{
	BLOCK_CACHE *head = refHeadObj(offset, listHead);
	BLOCK_CACHE *cache;

	if (head == NULL) {
		printf("No list\n");
	}
	else {
		cache = head;
		do {
			printDump(cache);
			printf("\n");
			cache = refNextList(offset, (OBJ_LIST*) ((uint) cache + offset));
		} while (cache != head);
	}
}

static void dump()
{
	// Xg_v
	printf("Search list\n");
	listDump(seqHead, OFFSETOF(BLOCK_CACHE, seqList));
	printf("\n");

	// XVXg_v
	printf("Update list\n");
	listDump(updateHead, OFFSETOF(BLOCK_CACHE, updateList));
	printf("\n");
	
	// ߂҂Xg_v
	printf("Write list\n");
	listDump(writeHead, OFFSETOF(BLOCK_CACHE, writeList));
	printf("\n");

	// ߂҂foCXinodeԍEZN^[ԍXg_v
	printf("Write sequential list\n");
	listDump(writeSeqHead, OFFSETOF(BLOCK_CACHE, writeSeqList));
	printf("\n");
}

static void dumpUio(struct uio *uio)
{
	struct iovec *iovec = uio->uio_iov;
	int i;

	printf("%s\n", (uio->uio_rw == UIO_READ) ? "UIO_READ" : "UIO_WRITE");
	printf("%d", uio->uio_offset);
	for (i = 0; i < uio->uio_iovcnt; ++i) {
		printf(" %d-%d", iovec[i].iov_base, iovec[i].iov_len);
	}
	printf("\n");
}

static int TASK_WAIT;
STATIC void delayWriteTask();

static void addToSchedule(int *dummy, int dummy2)
{
/*	static int cnt = 0;
	
	if (++cnt < 4) {
		return;
	}
	cnt = 0;
	delayWriteTask();
*/}

static int ioFunc(dev_t dev, struct uio *uio, int ioflag)
{
	dumpUio(uio);
	
	return 0;
}

static dev_t getDevInfo(int dummy)
{
	device.disk.sectors = 0xffffffff;
	device.disk.beginSector = 0;
	device.si_devsw = &si_devsw;
	si_devsw.d_read = ioFunc;
	si_devsw.d_write = ioFunc;
	return &device;
}

static void waitFlag(
	const int waitValue,	// ҂̒l
	int *flag)
{
	/*
	 * RpCōœKĖ[vȂ悤gotog
	 */
LOOP:
	if (*flag == waitValue) {
		goto LOOP;
	}
}

//============================ rbg}bv ==================================================

typedef struct {
	uchar sectorNum;	// LbVΔԍ
	uchar sectorCnt;	// LbVZN^[
	uchar sectorHigh;	// LbVŏʃZN^[ԍ
} CACHE_SECTOR;

static const CACHE_SECTOR bitmapSeq[] = {
	{0,0,0},{0,1,1},{1,1,2},{0,2,2},{2,1,3},{0,1,3},{1,2,3},{0,3,3},{3,1,4},{0,1,4},{1,1,4},{0,2,4},{2,2,4},{0,1,4},{1,3,4},{0,4,4},
	{4,1,5},{0,1,5},{1,1,5},{0,2,5},{2,1,5},{0,1,5},{1,2,5},{0,3,5},{3,2,5},{0,1,5},{1,1,5},{0,2,5},{2,3,5},{0,1,5},{1,4,5},{0,5,5},
	{5,1,6},{0,1,6},{1,1,6},{0,2,6},{2,1,6},{0,1,6},{1,2,6},{0,3,6},{3,1,6},{0,1,6},{1,1,6},{0,2,6},{2,2,6},{0,1,6},{1,3,6},{0,4,6},
	{4,2,6},{0,1,6},{1,1,6},{0,2,6},{2,1,6},{0,1,6},{1,2,6},{0,3,6},{3,3,6},{0,1,6},{1,1,6},{0,2,6},{2,4,6},{0,1,6},{1,5,6},{0,6,6},
	{6,1,7},{0,1,7},{1,1,7},{0,2,7},{2,1,7},{0,1,7},{1,2,7},{0,3,7},{3,1,7},{0,1,7},{1,1,7},{0,2,7},{2,2,7},{0,1,7},{1,3,7},{0,4,7},
	{4,1,7},{0,1,7},{1,1,7},{0,2,7},{2,1,7},{0,1,7},{1,2,7},{0,3,7},{3,2,7},{0,1,7},{1,1,7},{0,2,7},{2,3,7},{0,1,7},{1,4,7},{0,5,7},
	{5,2,7},{0,1,7},{1,1,7},{0,2,7},{2,1,7},{0,1,7},{1,2,7},{0,3,7},{3,1,7},{0,1,7},{1,1,7},{0,2,7},{2,2,7},{0,1,7},{1,3,7},{0,4,7},
	{4,3,7},{0,1,7},{1,1,7},{0,2,7},{2,1,7},{0,1,7},{1,2,7},{0,3,7},{3,4,7},{0,1,7},{1,1,7},{0,2,7},{2,5,7},{0,1,7},{1,6,7},{0,7,7},
	{7,1,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,1,8},{0,1,8},{1,1,8},{0,2,8},{2,2,8},{0,1,8},{1,3,8},{0,4,8},
	{4,1,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,2,8},{0,1,8},{1,1,8},{0,2,8},{2,3,8},{0,1,8},{1,4,8},{0,5,8},
	{5,1,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,1,8},{0,1,8},{1,1,8},{0,2,8},{2,2,8},{0,1,8},{1,3,8},{0,4,8},
	{4,2,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,3,8},{0,1,8},{1,1,8},{0,2,8},{2,4,8},{0,1,8},{1,5,8},{0,6,8},
	{6,2,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,1,8},{0,1,8},{1,1,8},{0,2,8},{2,2,8},{0,1,8},{1,3,8},{0,4,8},
	{4,1,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,2,8},{0,1,8},{1,1,8},{0,2,8},{2,3,8},{0,1,8},{1,4,8},{0,5,8},
	{5,3,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,1,8},{0,1,8},{1,1,8},{0,2,8},{2,2,8},{0,1,8},{1,3,8},{0,4,8},
	{4,4,8},{0,1,8},{1,1,8},{0,2,8},{2,1,8},{0,1,8},{1,2,8},{0,3,8},{3,5,8},{0,1,8},{1,1,8},{0,2,8},{2,6,8},{0,1,8},{1,7,8},{0,8,8}
};

static const u_int8_t cacheBitmap[8][9] = {
	{0, 0x1,  0x3,  0x7,  0xf,  0x1f, 0x3f, 0x7f, 0xff},
	{0, 0x2,  0x6,  0xe,  0x1e, 0x3e, 0x7e, 0xfe, 0xfe},
	{0, 0x4,  0xc,  0x1c, 0x3c, 0x7c, 0xfc, 0xfc, 0xfc},
	{0, 0x8,  0x18, 0x38, 0x78, 0xf8, 0xf8, 0xf8, 0xf8},
	{0, 0x10, 0x30, 0x70, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0},
	{0, 0x20, 0x60, 0xe0, 0xe0, 0xe0, 0xe0, 0xe0, 0xe0},
	{0, 0x40, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0},
	{0, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80}
};

/*
 * ݂ĂLbVZN^[擾
 *sӁtԉʂ̃rbgʂɑĂrbgʒuƃrbgԂ
 * return : CACHE_SECTOR
 */
STATIC INLINE CACHE_SECTOR getExistCacheSector(
	const u_int8_t bitmap)
{
	return bitmapSeq[bitmap];
}

/*
 * ݂ĂLbVZN^[擾
 * return : CACHE_SECTOR
 */
STATIC INLINE uchar getExistBitmap(
	const int sectorNum,	// LbV΃ZN^[ԍ
	const int sectorCnt)	// LbVZN^[
{
	return cacheBitmap[sectorNum][sectorCnt];
}

/*
 *sXt
 * LbVZN^[LbVrbg}bvɒǉ
 */
STATIC INLINE void addCacheSector(
	const int sectorNum,	// LbV΃ZN^[ԍ
	const int sectorCnt,	// LbVZN^[
	BLOCK_CACHE *m_blkCache)
{
	ASSERT(sectorNum < 8);
	ASSERT(sectorCnt <= 8);

	m_blkCache->bitmap |= cacheBitmap[sectorNum][sectorCnt];
}

/*
 *sXt
 * LbVZN^[XVrbg}bvɒǉ
 */
STATIC INLINE void setDirtySector(
	const int sectorNum,	// LbV΃ZN^[ԍ
	const int sectorCnt,	// LbVZN^[
	BLOCK_CACHE *m_blkCache)
{
	m_blkCache->dirtyBitmap |= cacheBitmap[sectorNum][sectorCnt];
}

/*
 *sXt
 * XVrbg}bvZbg
 */
STATIC INLINE void resetDirtyBitmap(
	BLOCK_CACHE *m_blkCache)
{
	m_blkCache->dirtyBitmap = 0;
}

/*
 *sQt
 * XVrbg}bvZbgĂ邩H
 * return : YES or NO
 */
STATIC INLINE int isDirtyCache(
	const BLOCK_CACHE *blkCache)
{
	return (blkCache->dirtyBitmap != 0) ? YES : NO;
}

/*
 *sXt
 * XVrbg}bv߂rbg}bvɉ
 */
STATIC INLINE void setWriteBitmap(
	BLOCK_CACHE *m_blkCache)
{
	m_blkCache->writeBitmap |= m_blkCache->dirtyBitmap;
}

/*
 *sXt
 * XVrbg}bv߂rbg}bvɉ
 */
STATIC INLINE void resetWriteBitmap(
	BLOCK_CACHE *m_blkCache)
{
	m_blkCache->writeBitmap = 0;
}

//============================ foCXhn ==================================================

// foCXIOpiovec
static struct iovec *iovRead;
static struct iovec *iovWrite;

/*
 * LbVUIOɐݒ肷
 * return : error number
 */
STATIC INLINE void setUio(
	const int iovecCnt,
	const int uioMode,
	const uint topSector,	// 擪ZN^[ԍ
	struct iovec *iovec,
	struct uio *m_uio)
{
	m_uio->uio_iovcnt = iovecCnt;
	m_uio->uio_rw = uioMode;
	m_uio->uio_iov = iovec;
	m_uio->uio_offset = topSector * CACHE_SECTOR_SIZE;
}

/*
 * foCXLbVɓǂݍށB
 * return : error number
 */
STATIC int readDevice(
	const int sectorNum,		// LbV΃ZN^[ԍ
	const int sectorCnt,		// ǂݍ݃ZN^[
	BLOCK_CACHE *m_blkCache)
{
	dev_t dev = getDevInfo(m_blkCache->din);
	uchar readBitmap = getExistBitmap(sectorNum, sectorCnt);
	CACHE_SECTOR cacheSector;
	
	for (readBitmap &= ~m_blkCache->bitmap; readBitmap != 0; readBitmap &= ~getExistBitmap(cacheSector.sectorNum, cacheSector.sectorCnt)) {
		struct uio uio;
		int error;

		cacheSector = getExistCacheSector(readBitmap);
		iovRead[0].iov_base = m_blkCache->cache + cacheSector.sectorNum * CACHE_SECTOR_SIZE;
		iovRead[0].iov_len = cacheSector.sectorCnt * CACHE_SECTOR_SIZE;

		// foCXIO
		setUio(1, UIO_READ, m_blkCache->sector + cacheSector.sectorNum + dev->disk.beginSector, iovRead, &uio);
		error = dev->si_devsw->d_read(dev, &uio, 0);
		if (error != NOERR) {
			return error;
		}
	}

	addCacheSector(sectorNum, sectorCnt, m_blkCache);

	return NOERR;
}

/*
 * LbVfoCXɏ߂B
 * return : error number
 */
STATIC int writeDevice(
	BLOCK_CACHE *i_blkCache)	// IO擪ubNLbV
{
	dev_t dev = getDevInfo(i_blkCache->din);
	struct uio uio;
	BLOCK_CACHE *blkCache;
	uint topSector;				// 擪ZN^[
	int c_iovec;

	/*
	 * LbVNiovWriteݒ肵IO
	 *sӁtLbVN̓ZN^[AĂ邱ƂOɂĂ
	 */
	c_iovec = 0;
	for (blkCache = i_blkCache; blkCache != NULL; blkCache = blkCache->cacheNext) {
		uchar bitmap;
		CACHE_SECTOR cacheSector;

		for (bitmap = blkCache->writeBitmap; bitmap != 0; bitmap &= ~getExistBitmap(cacheSector.sectorNum, cacheSector.sectorCnt)) {
			cacheSector = getExistCacheSector(bitmap);
			iovWrite[c_iovec].iov_base = blkCache->cache + cacheSector.sectorNum * CACHE_SECTOR_SIZE;
			iovWrite[c_iovec].iov_len = cacheSector.sectorCnt * CACHE_SECTOR_SIZE;
			++c_iovec;

			if (1 == c_iovec) {
				topSector = blkCache->sector + cacheSector.sectorNum;
			}

			if ((IOVEC_MAX <= c_iovec) || (cacheSector.sectorNum + cacheSector.sectorCnt < CACHE_SECTORS)){
				int error;

				// foCXIO
				setUio(c_iovec, UIO_WRITE, topSector + dev->disk.beginSector, iovWrite, &uio);
				error = dev->si_devsw->d_write(dev, &uio, 0);
				if (error != NOERR) {
					return error;
				}
				c_iovec = 0;
			}
		}
	}

	if (0 < c_iovec) {
		setUio(c_iovec, UIO_WRITE, topSector + dev->disk.beginSector, iovWrite, &uio);
		return dev->si_devsw->d_write(dev, &uio, 0);
	}
	else {
		return NOERR;
	}
}

//============================ x =======================================================

static PROC *writeTask;				// xݗp^XN

/*
 * xݑ҂tOZbg
 */
STATIC INLINE void setWriteBusy(
	BLOCK_CACHE *m_blkCache)
{
	m_blkCache->writeBusy = 1;
}

/*
 * xݑ҂tOZbg
 */
STATIC INLINE void resetWriteBusy(
	BLOCK_CACHE *m_blkCache)
{
	m_blkCache->writeBusy = 0;
}

/*
 * xݑ҂H
 * return : YES or NO
 */
STATIC INLINE int isWriteBusy(
	const BLOCK_CACHE *blkCache)
{
	return (blkCache->writeBusy == 1) ? YES : NO;
}

/*
 * xݑ҂
 */
STATIC INLINE void waitWriteBusy(
	BLOCK_CACHE *blkCache)
{
	if (isWriteBusy(blkCache) == YES) {
		// x݃^XNN
		addToSchedule(writeTask, TASK_WAIT);
		waitFlag(1, &blkCache->writeBusy);
	}
}

/*
 *sӁtwriteLockŃbNĂ钆Ă΂邱
 * ߂҂zXgNH
 * return : YES or NO
 */
STATIC INLINE int isWriteListLink(
	BLOCK_CACHE *blkCache)
{
	if (refNextList(OFFSETOF(BLOCK_CACHE, writeList), &blkCache->writeList) != blkCache) {
		return YES;
	}
	else if (refHeadObj(OFFSETOF(BLOCK_CACHE, writeList), writeHead) == blkCache) {
		return YES;
	}
	return NO;
}

/*
 *sXt
 * ZN^[qĂLbVN
 * return : q : TRUE, qȂ : FALSE
 */
STATIC INLINE int linkCache(
	BLOCK_CACHE *m_dstCache,
	BLOCK_CACHE *m_srcCache)		// ڑLbV
{
	CACHE_SECTOR cacheDst;
	CACHE_SECTOR cacheSrc;

	cacheDst = getExistCacheSector(m_dstCache->writeBitmap);
    cacheSrc = getExistCacheSector(m_srcCache->writeBitmap);

	ASSERT(0 < cacheDst.sectorCnt);

	if (m_dstCache->sector + cacheDst.sectorHigh == m_srcCache->sector + cacheSrc.sectorNum) {
		m_dstCache->cacheNext = m_srcCache;

		return TRUE;
	}
	else {
		return FALSE;
	}
}

/*
 *sXt
 *sӁtwriteLockŃbNĂ钆Ă΂邱
 * xݑ҂foCXinodeԍEZN^[ԍXgɉ
 */
STATIC void addWriteSeq(
	BLOCK_CACHE *m_blkCache)
{
	BLOCK_CACHE *head = refHeadObj(OFFSETOF(BLOCK_CACHE, writeSeqList), writeSeqHead);
	
	if (head == NULL) {
		insertHeadCrclList(&writeSeqHead, &m_blkCache->writeSeqList);
	}
	else if (m_blkCache->din < head->din) {
		insertHeadCrclList(&writeSeqHead, &m_blkCache->writeSeqList);
	}
	else if ((m_blkCache->din == head->din) && (m_blkCache->sector < head->sector)) {
		insertHeadCrclList(&writeSeqHead, &m_blkCache->writeSeqList);
	}
	else {
		BLOCK_CACHE *blkCache;

		for (blkCache = refNextList(OFFSETOF(BLOCK_CACHE, writeSeqList), &head->writeSeqList);
			blkCache != head;
			blkCache = refNextList(OFFSETOF(BLOCK_CACHE, writeSeqList), &blkCache->writeSeqList)) {
			
			ASSERT((m_blkCache->din != blkCache->din) || (m_blkCache->sector != blkCache->sector));
			
			if (m_blkCache->din < blkCache->din) {
				break;
			}
			else if ((m_blkCache->din == head->din) && (m_blkCache->sector < blkCache->sector)) {
				break;
			}
		}
		insertPrevList(&blkCache->writeSeqList, &m_blkCache->writeSeqList);
	}
}

/*
 *sXt
 * xݑ҂Xgɉ
 */
STATIC void addWriteWait(
	BLOCK_CACHE *m_blkCache)
{
	enter_spinlock(&writeLock);
	{
		if (isWriteListLink(m_blkCache) == YES) {
			removeOutHeadList(&writeHead, &m_blkCache->writeList);
		}
		else {
			waitWriteBusy(m_blkCache);

			addWriteSeq(m_blkCache);
			setWriteBusy(m_blkCache);
		}

		// ߂rbg}bṽZbg
		setWriteBitmap(m_blkCache);

		insertHeadCrclList(&writeHead, &m_blkCache->writeList);
	}
	exit_spinlock(&writeLock);

	// x݃^XNN
	addToSchedule(writeTask, TASK_WAIT);
}

/*
 *sXt
 * ߂҂zXg̍Ō̃LbVƃZN^[qĂLbVNĎo
 * return : ߂҂LbV or NULL
 */
STATIC BLOCK_CACHE *getWriteWait()
{
	BLOCK_CACHE *linkTop;

	enter_spinlock(&writeLock);
	{
		BLOCK_CACHE *writeCache = getHeadPrevList(OFFSETOF(BLOCK_CACHE, writeList), &writeHead);
		BLOCK_CACHE *linkEnd;
		BLOCK_CACHE *limit;
		BLOCK_CACHE *blkCache;
		BLOCK_CACHE *next;
		BLOCK_CACHE *prev;

		if (writeCache == NULL) {
			exit_spinlock(&writeLock);
			return NULL;
		}

		linkTop = writeCache;
		limit = refHeadObj(OFFSETOF(BLOCK_CACHE, writeSeqList), writeSeqHead);
		for (blkCache = refPrevList(OFFSETOF(BLOCK_CACHE, writeSeqList), &writeCache->writeSeqList);; blkCache = prev) {
			prev = refPrevList(OFFSETOF(BLOCK_CACHE, writeSeqList), &blkCache->writeSeqList);
			
			if ((blkCache->din == linkTop->din) && (linkCache(blkCache, linkTop) == TRUE)) {
				removeOutHeadList(&writeSeqHead, &blkCache->writeSeqList);
				removeOutHeadList(&writeHead, &blkCache->writeList);
				linkTop = blkCache;
			}
			else {
				break;
			}

			if (blkCache == limit) {
				break;
			}
		}

		linkEnd = writeCache;
		limit = refHeadPrevList(OFFSETOF(BLOCK_CACHE, writeSeqList), writeSeqHead);
		for (blkCache = refNextList(OFFSETOF(BLOCK_CACHE, writeSeqList), &writeCache->writeSeqList);; blkCache = next) {
			next = refNextList(OFFSETOF(BLOCK_CACHE, writeSeqList), &blkCache->writeSeqList);

			if ((blkCache->din == linkEnd->din) && (linkCache(linkEnd, blkCache) == TRUE)) {
				removeOutHeadList(&writeSeqHead, &blkCache->writeSeqList);
				removeOutHeadList(&writeHead, &blkCache->writeList);
				linkEnd = blkCache;
			}
			else {
				break;
			}
			
			if (blkCache == limit) {
				break;
			}
		}
		
		removeOutHeadList(&writeSeqHead, &writeCache->writeSeqList);
	}
	exit_spinlock(&writeLock);

	return linkTop;
}

/*
 *sXt
 * xݏI
 */
STATIC void endWriteCache(
	BLOCK_CACHE *m_blkCache)
{
	BLOCK_CACHE *blkCache;
	BLOCK_CACHE *next;

	for (blkCache = m_blkCache; blkCache != NULL; blkCache = next) {
		next = blkCache->cacheNext;
		blkCache->cacheNext = NULL;
		resetWriteBitmap(blkCache);
		resetWriteBusy(blkCache);	// rW[tO͑̑SڂZbgɃZbg
	}
}

/*
 *************************************
 */
STATIC void delayWriteTask()
{
	for (;;) {
		// ݑ҂o
		BLOCK_CACHE *blkCache = getWriteWait();

		if (blkCache != NULL){
			if (writeDevice(blkCache) != NOERR){
				printk("Device error in delay write function!\n");
			}
			endWriteCache(blkCache);
		}
		else {
			break;
		}
	}
}

/*
 * xݑ҂p[^̏
 */
STATIC void initWriteParam(
	BLOCK_CACHE *m_blkCache)
{
	initCirculList(&m_blkCache->writeList);
	initCirculList(&m_blkCache->writeSeqList);
	m_blkCache->cacheNext = NULL;
	m_blkCache->writeBusy = 0;
}

//============================ LbV ==================================================

/*
 * ubNLbV̏
 */
STATIC void initCache(
	BLOCK_CACHE *m_blkCache)
{
	char *cache = m_blkCache->cache;

	memset(m_blkCache, 0, sizeof(*m_blkCache));
	m_blkCache->cache = cache;
	initCirculList(&m_blkCache->seqList);
	initCirculList(&m_blkCache->updateList);
	
	initWriteParam(m_blkCache);
}

/*
 *sXt
 * LbVfBXNɏ߂
 */
STATIC void writeCache(
	BLOCK_CACHE *m_blkCache)
{
	// xݑ҂Nɐڑ
	addWriteWait(m_blkCache);

	resetDirtyBitmap(m_blkCache);
}

/*
 *sXt
 * ŐVLbVɂ
 */
STATIC void updateCache(
	BLOCK_CACHE *m_blkCache)
{
	enum {
		WRITE_BACK_COUNT = 4,	// fBXNɏ߂LbV̍ŐV̏
	};
	BLOCK_CACHE *new;
	int i;

	// ŐVɍXV
	removeOutHeadList(&updateHead, &m_blkCache->updateList);
	insertHeadCrclList(&updateHead, &m_blkCache->updateList);
	
	// ݂LbVfBXNɏ߂
	new = m_blkCache;
	for (i = 1; i < WRITE_BACK_COUNT; ++i) {
		new = refNextList(OFFSETOF(BLOCK_CACHE, updateList), &new->updateList);
	}
	if (isDirtyCache(new) == YES) {
		writeCache(new);
	}
}

/*
 *sQt
 * ΈɎ̃ubNLbVԂ
 * return : ubNLbV or NULL
 */
STATIC BLOCK_CACHE *searchCache(
	const int din,			// foCXinode
	const uint sector,		// ZN^[ԍ
	BLOCK_CACHE **o_next)	// ̃ubNLbV
{
	BLOCK_CACHE *head = refHeadObj(OFFSETOF(BLOCK_CACHE, seqList), seqHead);
	BLOCK_CACHE *find;
	BLOCK_CACHE *next;

	if (head == NULL) {
		*o_next = NULL;
		return NULL;
	}

	/*
	 * NQƂăfoCXinodeƃZN^[ԍvLbV擾
	 */
	find = NULL;	// ꂽLbV
	*o_next = NULL;
	next = head;
	do {
		if (din < next->din) {
			*o_next = next;
			break;
		}
		
		if (next->din == din) {
			if (sector < next->sector) {
				*o_next = next;
				break;
			}
			else if ((next->sector <= sector) && (sector < next->sector + CACHE_SECTORS)) {
				find = next;
				break;
			}
		}
		
		next = refNextList(OFFSETOF(BLOCK_CACHE, seqList), &next->seqList);
	} while (next != head);
	
	return find;
}

/*
 *sQt
 * 
 * return : ubNLbV or NULL
 */
STATIC BLOCK_CACHE *findCache(
	const int din,			// foCXinode
	const uint sector)		// ZN^[ԍ
{
	BLOCK_CACHE *next;

	return searchCache(din, sector, &next);
}

/*
 * LbVNɂȂ
 * N̓foCXinodeEubNɃ\[g悤ɂȂ
 */
STATIC void addCache(
	BLOCK_CACHE *m_blkCache,
	BLOCK_CACHE *next)
{
	if (next == NULL) {
		BLOCK_CACHE *prev = refHeadPrevList(OFFSETOF(BLOCK_CACHE, seqList), seqHead);
		
		if (prev == NULL) {				// NɉȂ
			insertHeadCrclList(&seqHead, &m_blkCache->seqList);
		}
		else {
			insertNextList(&prev->seqList, &m_blkCache->seqList);
		}
	}
	else {
		BLOCK_CACHE *head = refHeadObj(OFFSETOF(BLOCK_CACHE, seqList), seqHead);

		if (next == head) {
			insertHeadCrclList(&seqHead, &m_blkCache->seqList);
		}
		else {
			insertPrevList(&next->seqList, &m_blkCache->seqList);
		}
	}
}

/*
 *sXt
 * VɃLbV擾
 * return : LbV
 */
STATIC BLOCK_CACHE *getCache(
	const int din,			// foCXinode
	const uint i_sector)	// ZN^[ԍ
{
	const uint sector = ROUNDDOWN(i_sector, CACHE_SECTORS);
	BLOCK_CACHE *blkCache;
	BLOCK_CACHE *next;
	
	// ԌÂLbVo
	blkCache = getHeadPrevList(OFFSETOF(BLOCK_CACHE, updateList), &updateHead);

	// ԌÂLbV͏߂Ă͂
	ASSERT((isWriteBusy(blkCache) == YES) || (isDirtyCache(blkCache) == NO));
	
	// ݏI҂
	waitWriteBusy(blkCache);
	
	// N苎
	removeOutHeadList(&seqHead, &blkCache->seqList);

	// 
	initCache(blkCache);
	blkCache->din = din;
	blkCache->sector = sector;

	// Nɐڑ
	// ĂȂɌĂ΂̂Ō͂͂Ȃ
	ASSERT(searchCache(din, sector, &next) == NULL);
//	searchCache(din, sector, &next);
	addCache(blkCache, next);

	return blkCache;
}

//================================== PUBLIC =============================================

/*
 * ݂LbVubNׂĒx݂ɑB
 */
void writeBackAllCache()
{
	enter_spinlock(&cacheLock);
	{
		BLOCK_CACHE *head = refHeadObj(OFFSETOF(BLOCK_CACHE, seqList), seqHead);
		BLOCK_CACHE *blkCache;
		
		blkCache = head;
		do {
			if ((isWriteBusy(blkCache) == NO) && (isDirtyCache(blkCache) == YES)) {
				writeCache(blkCache);
			}

			blkCache = refNextList(OFFSETOF(BLOCK_CACHE, seqList), &blkCache->seqList);
		} while (blkCache != head);
		
		while (refHeadObj(OFFSETOF(BLOCK_CACHE, writeList), writeHead) != NULL) {
			wait_task();
		}
	}
	exit_spinlock(&cacheLock);
}

/*
 *sӁt荞݃nȟĂяos
 *streadCache()𕡐LbVǂݍ݂ɑΉB
 * return : error number
 */
int read_cache(
	const int din,				// foCXinode
	void *i_readBuf,			// readobt@
	const size_t i_sectors,		// readZN^[
	const size_t i_begin)		// readJnZN^[
{
	char *readBuf = i_readBuf;
	dev_t dev = getDevInfo(din);
	uint sector;
	uint sectorLast;
	uint sectorBound;

	// _fBXÑ`FbN
	if (dev->disk.sectors < i_begin + i_sectors){
		return -EIO;
	}

	/*
	 * LbViPAGE_SIZEjPʂIO
	 * sector : LbVPʂ̓ǂݍ݊JnZN^[ԍ
	 * sectorLast : XgIOZN^[ԍ
	 * sectorBound : PAGE_SIZEẼZN^[ԍ
	 * unreadCnt : foCX疢ǂݍ݃LbV
	 * unreadCache : foCX疢ǂݍ݃LbV擪
	 */
	sectorLast = i_begin + i_sectors;
	sectorBound = ROUNDDOWN(i_begin, CACHE_SECTORS);
	for (sector = i_begin; sector < sectorLast;) {
		BLOCK_CACHE *blkCache;	// IOLbV
		uint sectorNext;		// PAGE_SIZEẼZN^[ԍ
		uint sectorNum;			// obt@Rs[ZN^[
		int error;

		enter_spinlock(&cacheLock);
		{
			sectorNext = ROUNDDOWN(sector + CACHE_SECTORS, CACHE_SECTORS);

			// hnZN^[vZ
			if (sectorLast < sectorNext) {
				sectorNum = sectorLast - sector;
			}
			else {
				sectorNum = sectorNext - sector;
			}

			// LbV
			blkCache = findCache(din, sector);
			if (blkCache == NULL) {
				// LbVɂȂΐVɊ蓖Ă
				blkCache = getCache(din, sector);
			}

			// fBXNǂݍ܂ĂȂZN^[ǂݍ
			error = readDevice(sector - sectorBound, sectorNum, blkCache);
			if (error != NOERR) {
				// readG[̓fBXv[̂
				printk("readCache() error! : %d\n", error);
			}

			// LbVɃRs[
			memcpy(readBuf, blkCache->cache + (sector - sectorBound) * CACHE_SECTOR_SIZE, sectorNum * CACHE_SECTOR_SIZE);
			updateCache(blkCache);
		}
		exit_spinlock(&cacheLock);

		readBuf += sectorNum * CACHE_SECTOR_SIZE;
		sector = sectorNext;
		sectorBound = sectorNext;
	}

	return NOERR;
}

/*
 *sӁt 荞݃nȟĂяos
 * return : error number
 */
int write_cache(
	const int din,				// foCXinode
	void *i_writeBuf,			// writeobt@
	const size_t i_sectors, 	// writeZN^[
	const size_t i_begin)		// writeJnZN^[
{
	char *writeBuf = i_writeBuf;
	dev_t dev = getDevInfo(din);
	uint sector;
	uint sectorLast;
	uint sectorBound;

	// _fBXÑ`FbN
	if (dev->disk.sectors < i_begin + i_sectors){
		return -EIO;
	}

	/*
	 * LbViPAGE_SIZEjPʂIO
	 * sector : LbVPʂ̓ǂݍ݊JnZN^[ԍ
	 * sectorLast : XgIOZN^[ԍ
	 * sectorBound : PAGE_SIZEẼZN^[ԍ
	 */
	sectorLast = i_begin + i_sectors;
	sectorBound = ROUNDDOWN(i_begin, CACHE_SECTORS);
	for (sector = i_begin; sector < sectorLast;) {
		BLOCK_CACHE *blkCache;	// IOLbV
		uint sectorNext;		// PAGE_SIZEẼZN^[ԍ
		uint sectorNum;			// obt@Rs[ZN^[

		enter_spinlock(&cacheLock);
		{
			sectorNext = ROUNDDOWN(sector + CACHE_SECTORS, CACHE_SECTORS);

			// hnZN^[vZ
			if (sectorLast < sectorNext) {
				sectorNum = sectorLast - sector;
			}
			else {
				sectorNum = sectorNext - sector;
			}

			// LbV
			blkCache = findCache(din, sector);
			if (blkCache == NULL) {
				// LbVɂȂΐVɊ蓖Ă
				blkCache = getCache(din, sector);
			}

			// LbVɃRs[
			memcpy(blkCache->cache + (sector - sectorBound) * CACHE_SECTOR_SIZE, writeBuf, sectorNum * CACHE_SECTOR_SIZE);
			setDirtySector(sector - sectorBound, sectorNum, blkCache);
			addCacheSector(sector - sectorBound, sectorNum, blkCache);
			updateCache(blkCache);
		}
		exit_spinlock(&cacheLock);

		writeBuf += sectorNum * CACHE_SECTOR_SIZE;
		sector = sectorNext;
		sectorBound = sectorNext;
	}

	return NOERR;
}

/*
 *****************************
 */
int initBlockCache()
{
	enum {
		IOVEC_MAX = 5,
	};
	int i;

	// ubNLbVCX^X̏
	for (i = 0; i < CACHE_NUM_MAX; ++i){
		initCache(&blkCacheInstance[i]);
		
		blkCacheInstance[i].cache = 0;
		
		// ŐVNɂȂ
		insertHeadCrclList(&updateHead, &blkCacheInstance[i].updateList);
	}

	iovRead = kmalloc(sizeof(*iovRead) * IOVEC_MAX);
	if (iovRead == NULL) {
		return -ENOMEM;
	}
	iovWrite = kmalloc(sizeof(*iovWrite) * IOVEC_MAX);
	if (iovWrite == NULL) {
		return -ENOMEM;
	}

	return NOERR;
}

/******************************************************************************************/

int main(int argc, char *argv[])
{
	initBlockCache();

	for (;;) {
		int func = 1;
		int din = 0;
		int sectorCnt;
		int sector;

//		scanf("%d %d %d %d", &func , &din, &sector, &sectorCnt);
		scanf("%d %d %d", &func, &sector, &sectorCnt);

		switch (func) {
		case 0:
			read_cache(din, NULL, sectorCnt, sector);
			dump();
			break;
		case 1:
			write_cache(din, NULL, sectorCnt, sector);
			dump();
			break;
		case 2:
			delayWriteTask();
			dump();
		default:
			break;
		}
	}

	return 0;
}
