/*
 * ext2_test.c
 *
 * EXT2eXgp
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


enum {
	NOERR = 0,
	ENOMEM,
	EIO,
	EFBIG,

};
#define STATIC static
#define INLINE inline
#define kmalloc malloc
#define kfree free
#define ROUNDUP(a, b) (((uint)(a) + (b) - 1) / (b) * (b))
#define ROUNDDOWN(a, b) ((uint)(a) / (b) * (b))
#define ROUNDUP_DIV(a, b) (((uint)(a) + (b) - 1) / (b))
#define ASSERT(bool)
//#define ASSERT(bool) if ((bool) == 0){							\
//	printf("Assertion! : %s line=%d \n",__FILE__, __LINE__); 	\
//}
#define WARNING(bool) if ((bool) == 0){							\
	printf("Warning! : %s line=%d \n",__FILE__, __LINE__);		\
}
#define ARRAY_LENGTH(x)	(sizeof(x) / sizeof(x[0]))

#define SET_MEMPOOL(name, max, memSize) 									\
	char mempoolArea[max * sizeof(void*) + sizeof(struct mempool)];			\
	struct mempool *name = (struct mempool*)mempoolArea;					\
	memset(mempoolArea, 0, max * sizeof(void*) + sizeof(struct mempool));	\
	name->maxCount = max;													\
	name->size = memSize;
typedef unsigned char uchar;
struct mempool {
	size_t	size;			// 蓖Ă郁[TCY
	int		maxCount;		// ő僁[JE^iݒj
	int		allocCount;		// JE^iVȊɃvXPj
	int		currentCount;	// [JE^ilOAIɑj
	void	*pointer[0];	// ۑ[|C^
};

/******************* fBXNG~[V *********************************/

enum {
	SECTOR_BYTES	= 12,
	SECTOR_PER_INT	= SECTOR_BYTES / sizeof(int),
	SECTOR_MAX		= 60,	// ZN^[
};

typedef union {
	uchar byte[SECTOR_BYTES];
} SECTOR;

static SECTOR sector[SECTOR_MAX];

/*
 * return : error number
 */
static int read_cache(int din, void *buf, size_t sectors, size_t begin, size_t *o_sectors)
{
/****************************************************************************************************************/
fprintf(stderr,"read_cache() begin=%d\n", begin);
/****************************************************************************************************************/
	if (SECTOR_MAX < begin + sectors) {
		return -EIO;
	}
	
	memcpy(buf, &sector[begin], sizeof(SECTOR) * sectors);
	
	return NOERR;
}

/*
 * return : error number
 */
static int write_cache(int din, void *buf, size_t sectors, size_t begin, size_t *o_sectors)
{
/****************************************************************************************************************/
fprintf(stderr,"write_cache() begin=%d\n", begin);
/****************************************************************************************************************/
	if (SECTOR_MAX < begin + sectors) {
		return -EIO;
	}
	
	memcpy(&sector[begin], buf, sizeof(SECTOR) * sectors);
	
	return NOERR;
}

/*
 * fBXNj^[
 */
static void moniterDisk()
{
	int i;

	for (i = 0; i < 60; ++i) {
		int j;
		if ((i % 4) == 0) {
			printf("\n");
		}
		printf("%02d ", i);
		for (j = 0; j < SECTOR_BYTES; ++j) {
			printf("%02x", sector[i].byte[j]);
		}
		printf(" ");
	}
	printf("\n");
}

static void initDisk()
{
	memset(sector, 0, sizeof(sector[SECTOR_MAX]));
}

//*******************************************************************

enum {
	INODE_DATA_INDIRECT1	= 2,		/* ԐڎQƃubNzlB */
	INODE_DATA_INDIRECT2	= 3,		/* dԐڎQƃubNzlB */
	INODE_DATA_INDIRECT3	= 4,		/* OdԐڎQƃubNzlB */
	INODE_DATA_BLOCK_NUM	= INODE_DATA_INDIRECT3 + 1,
};

#define EXT2_SECT_NUM(din)		1
#define EXT2_BLOCK_SIZE(din)	12

typedef struct{
	uint i_size;
	uint i_block[INODE_DATA_BLOCK_NUM];
}INODE;

STATIC INLINE uint blkToSect(
	int din, 
	uint blk)
{
	return blk * EXT2_SECT_NUM(din);
}

STATIC INLINE int is4Kbyte(int din)
{
	return 0;
}

//------------------  -------------------------------

static int getBlock(const int din, int *o_error)
{
	static int block = 0;
	
	return ++block;
}

static void initDirectBlock(const int, const int);
static void initIndirectBlock(const int, const int);
static void (*initBlock[])(const int, const int) = {
	&initDirectBlock,
	&initIndirectBlock,
	&initIndirectBlock,
	&initIndirectBlock
};

static void initDirectBlock(
	const int blk,		// ubNԍ
	const int node)
{
	static int num = 0;
	uchar block[SECTOR_BYTES * EXT2_SECT_NUM(0)];
	int i;
	int dummy;

	for (i = 0; i < SECTOR_BYTES; ++i) {
		block[i] = num++;
	}
	write_cache(0, block, EXT2_SECT_NUM(0), blkToSect(0, blk), &dummy);
}

static void initIndirectBlock(
	const int blk,		// ubNԍ
	const int node)
{
	uint block[SECTOR_BYTES * EXT2_SECT_NUM(0) / sizeof(uint)];
	int i;
	int dummy;

	for (i = 0; i < SECTOR_BYTES * EXT2_SECT_NUM(0) / sizeof(uint); ++i) {
		block[i] = getBlock(0, NULL);
		initBlock[node - 1](block[i], node - 1);
	}
	write_cache(0, block, EXT2_SECT_NUM(0), blkToSect(0, blk), &dummy);
}

/*
 * inodȅ
 */
static void initInode(INODE *m_inode)
{
	memset(m_inode, 0, sizeof(*m_inode));
}

/*
 * [̃j^[
 */
static void moniterMem(
	const uchar *buf,
	const size_t bytes)
{
	int i;

	for (i = 0; i < bytes; ++i) {
		if ((i % 16) == 0) {
			printf("\n");
		}
		printf("%02x", buf[i]);
	}
	printf("\n");
}

/******************* eXgR[h *************************************/

STATIC INLINE u_int32_t getSize(const INODE *inode)
{
	return inode->i_size;
}

//-------------------------------------- write data ------------------------------------------------

enum {
	INDIRECT_MAX = 4		// őԐڃubNm[h
};

typedef struct {
	u_int32_t blk[INDIRECT_MAX];	// ubNԍ
	void *blkBuf[INDIRECT_MAX];		// ubNobt@
	int upload[INDIRECT_MAX];		// XVtOF1 = XV or 0
} BLK_CACHE;

/*
 * ԐڃubNz񐔂擾
 */
STATIC INLINE uint getIndirectNode(
	const int din,
	const int node)		// Ԑڃm[h
{
	static uint blocks[2][4] =	{
		{INODE_DATA_INDIRECT1, 3,  3  * 3,  3  * 3  * 3},
		{INODE_DATA_INDIRECT1, 1024, 1024 * 1024, 1024 * 1024 * 1024}
	};

	return blocks[is4Kbyte(din)][node];
}

/*
 * f[^ubNLbV
 */
STATIC void initBlkCache(BLK_CACHE *m_blkCache)
{
	memset(m_blkCache, 0, sizeof(*m_blkCache));
}

/*
 * f[^ubNLbV
 * Ԑڃm[hƂɍXVǗ
 * return : ubNobt@ or NULL
 */
STATIC void *getBlkCache(
	const int din,
	const u_int32_t i_blk,
	const int node,
	BLK_CACHE *m_blkCache,
	int *o_error)
{
	size_t ioSize;

	ASSERT(node < INDIRECT_MAX);

	if (m_blkCache->blk[node] == i_blk) {
		*o_error = NOERR;
		return m_blkCache->blkBuf[node];
	}
	else {
		if (m_blkCache->blk[node] != 0) {
			if (m_blkCache->upload[node] == 1) {
				// ݂̃ubNLbV̓fBXNɏ߂
				*o_error = write_cache(din, m_blkCache->blkBuf[node], EXT2_SECT_NUM(din), blkToSect(din, m_blkCache->blk[node]), &ioSize);
				if (*o_error != NOERR) {
					return NULL;
				}
				ASSERT(EXT2_SECT_NUM(din) == ioSize);
			}
		}
		else {
			// obt@
			m_blkCache->blkBuf[node] = kmalloc(EXT2_BLOCK_SIZE(din));
			if (m_blkCache->blkBuf[node] == NULL) {
				*o_error = -ENOMEM;
				return NULL;
			}
		}

		// fBXNǂݍ
		*o_error = read_cache(din, m_blkCache->blkBuf[node], EXT2_SECT_NUM(din), blkToSect(din, i_blk), &ioSize);
		if (*o_error != NOERR) {
			return NULL;
		}
		ASSERT(EXT2_SECT_NUM(din) == ioSize);
		m_blkCache->blk[node] = i_blk;
		m_blkCache->upload[node] = 0;
		
		*o_error = NOERR;
		return m_blkCache->blkBuf[node];
	}
}

/*
 * f[^ubNLbVV蓖Ă
 * ubNXVtO𗧂Ă
 * return : ubNobt@ or NULL
 */
STATIC void *getNewBlkCache(
	const int din,
	const u_int32_t i_blk,
	const int node,
	BLK_CACHE *m_blkCache,
	int *o_error)
{
	size_t ioSize;

	ASSERT(node < INDIRECT_MAX);

	if (m_blkCache->blk[node] != 0) {
		if (m_blkCache->upload[node] == 1) {
			// ݂̃ubNLbV̓fBXNɏ߂
			*o_error = write_cache(din, m_blkCache->blkBuf[node], EXT2_SECT_NUM(din), blkToSect(din, m_blkCache->blk[node]), &ioSize);
			if (*o_error != NOERR) {
				return NULL;
			}
			ASSERT(EXT2_SECT_NUM(din) == ioSize);
		}
	}
	else {
		// ŏ̃obt@
		m_blkCache->blkBuf[node] = kmalloc(EXT2_BLOCK_SIZE(din));
		if (m_blkCache->blkBuf[node] == NULL) {
			*o_error = -ENOMEM;
			return NULL;
		}
	}

	memset(m_blkCache->blkBuf[node], 0, EXT2_BLOCK_SIZE(din));
	m_blkCache->blk[node] = i_blk;
	m_blkCache->upload[node] = 0;

	*o_error = NOERR;
	return m_blkCache->blkBuf[node];
}

/*
 * f[^ubNLbVJ
 * return : error number
 */
STATIC int freeBlkCache(
	const int din,
	BLK_CACHE *m_blkCache)
{
	size_t ioSize;
	int node;
	int error;
	
	for (node = 0; node < INDIRECT_MAX; ++node) {
		if (m_blkCache->blk[node] != 0) {
			if (m_blkCache->upload[node] == 1) {
				error = write_cache(din, m_blkCache->blkBuf[node], EXT2_SECT_NUM(din), blkToSect(din, m_blkCache->blk[node]), &ioSize);
				if (error != NOERR) {
					for (; node < INDIRECT_MAX; ++node) {
						kfree(m_blkCache->blkBuf[node]);
					}
					return error;
				}
				ASSERT(EXT2_SECT_NUM(din) == ioSize);
			}

			kfree(m_blkCache->blkBuf[node]);
		}
	}
	
	return NOERR;
}

/*
 * f[^ubN̍XVtO𗧂Ă
 *sӁtwriteŕKĂяo
 * return : error number
 */
STATIC void updateBlk(
	const int node,
	BLK_CACHE *m_blkCache)
{
	m_blkCache->upload[node] = 1;
}

/*
 * YItZbg̃ubNf[^擾
 *sӁtubN蓖Ă̂ŁAREADŃt@CTCY𒴂ČĂяoȂƁB
 * return : f[^|C^ or NULL
 */
STATIC char *getBlkData(
	const int din,				// device inode
	const int64_t i_offset,		// f[^oCgItZbg
	INODE *m_inode,				// inode
	BLK_CACHE *m_blkCache,		// f[^ubNLbV\
	int *o_error)
{
	uint blkNum = i_offset / EXT2_BLOCK_SIZE(din);	// ΃ubN
	int node;										// Ԑڃm[h

	// Ԑڃm[hƊYm[hł̑΃ubNZo
	node = 0;
	for (; getIndirectNode(din, node) <= blkNum; ++node) {
		if (node == INDIRECT_MAX) {
			*o_error = -EFBIG;
			return NULL;
		}
		blkNum -= getIndirectNode(din, node);
	}

	// ԐڃubNǂăf[^ubNobt@擾
	{
		u_int32_t *blkBuf;
		int blkArray;

		blkBuf = m_inode->i_block;
		blkArray = (node == 0) ? blkNum : INODE_DATA_INDIRECT1 + node - 1;
		for (;;) {
			if (blkBuf[blkArray] == 0) {
				// ubNVɊ蓖Ă
				blkBuf[blkArray] = getBlock(din, o_error);
				if (blkBuf[blkArray] == 0) {
					return NULL;
				}
				
				blkBuf = getNewBlkCache(din, blkBuf[blkArray], node, m_blkCache, o_error);
				if (*o_error != NOERR) {
					return NULL;
				}
			}
			else {
				blkBuf = getBlkCache(din, blkBuf[blkArray], node, m_blkCache, o_error);
				if (*o_error != NOERR) {
					return NULL;
				}
			}
			
			if (node == 0) {
				break;
			}

			if (--node == 0) {
				blkArray = blkNum;
			}
			else {
				blkArray = blkNum / getIndirectNode(din, node);
				blkNum %= getIndirectNode(din, node);
			}

			if (blkBuf[blkArray] == 0) {
				updateBlk(node + 1, m_blkCache);
			}
		}
		
		*o_error = NOERR;
		return (char*)blkBuf;
	}
}

/*
 * f[^ubNǂ
 * return : error number
 */
STATIC int readData(
	const int din,
	const u_int32_t i_offset,	// JnoCgItZbg
	const int64_t i_bytes,		// ǂݍ݃oCg
	char *m_readBuf,			// ǂݍ݃obt@
	INODE *inode)
{
	BLK_CACHE blkCache;
	char *readBuf = m_readBuf;
	char *blkData;
	int64_t offset;
	int64_t last;
	size_t readBytes;
	int error;

	initBlkCache(&blkCache);
	
	// ŏ̃ubNTCYǂ
	blkData = getBlkData(din, i_offset, inode, &blkCache, &error);
	if (error != NOERR) {
		goto END;
	}
	readBytes = ROUNDUP(i_offset, EXT2_BLOCK_SIZE(din)) - i_offset;
	if (i_bytes < readBytes) {
		readBytes = i_bytes;
	}
	memcpy(readBuf, blkData + (i_offset % EXT2_BLOCK_SIZE(din)), readBytes);
	readBuf += readBytes;

	// Ō̃ubNE܂œǂ
	last = ROUNDDOWN(i_offset + i_bytes, EXT2_BLOCK_SIZE(din));
	offset = ROUNDUP(i_offset, EXT2_BLOCK_SIZE(din));
	for (; offset < last; offset += EXT2_BLOCK_SIZE(din)) {
		blkData = getBlkData(din, offset, inode, &blkCache, &error);
		if (error != NOERR) {
			goto END;
		}
		memcpy(readBuf, blkData, EXT2_BLOCK_SIZE(din));
		readBuf += EXT2_BLOCK_SIZE(din);
	}
		
	if (offset < i_offset + i_bytes) {
		// Ō̃ubNTCY
		blkData = getBlkData(din, offset, inode, &blkCache, &error);
		if (error != NOERR) {
			goto END;
		}
		readBytes = i_offset + i_bytes - offset;
		memcpy(readBuf, blkData, readBytes);
	}
	
	error = NOERR;
END:
	freeBlkCache(din, &blkCache);
	
	return error;
}

/*
 * f[^ubNɏށB
 * return : error number
 */
STATIC int writeData(
	const int din,				// device s
	const u_int32_t i_offset,	// JnoCgItZbg
	const uint i_bytes,			// ݃oCg
	const void *i_writeBuf,		// ݃obt@
	INODE *m_inode)				// inode
{
	BLK_CACHE blkCache;
	const char *writeBuf = i_writeBuf;
	char *blkData;
	int64_t offset;
	int64_t last;
	size_t writeBytes;
	int error;

	initBlkCache(&blkCache);
	
	// ލŏ܂ł̖蓖ăubN蓖Ă
	if (ROUNDUP(getSize(m_inode), EXT2_BLOCK_SIZE(din)) < ROUNDDOWN(i_offset, EXT2_BLOCK_SIZE(din))) {
		last = ROUNDDOWN(i_offset, EXT2_BLOCK_SIZE(din));
		for (offset = ROUNDUP(getSize(m_inode), EXT2_BLOCK_SIZE(din)); offset < last; offset += EXT2_BLOCK_SIZE(din)) {
			getBlkData(din, offset, m_inode, &blkCache, &error);
			if (error != NOERR) {
				goto END;
			}
		}
	}
	
	// ŏ̃ubNTCY
	blkData = getBlkData(din, i_offset, m_inode, &blkCache, &error);
	if (error != NOERR) {
		goto END;
	}
	writeBytes = ROUNDUP(i_offset, EXT2_BLOCK_SIZE(din)) - i_offset;
	if (i_bytes < writeBytes) {
		writeBytes = i_bytes;
	}
	memcpy(blkData + (i_offset % EXT2_BLOCK_SIZE(din)), writeBuf, writeBytes);
	updateBlk(0, &blkCache);
	writeBuf += writeBytes;

	// Ō̃ubNE܂ŏ
	last = ROUNDDOWN(i_offset + i_bytes, EXT2_BLOCK_SIZE(din));
	offset = ROUNDUP(i_offset, EXT2_BLOCK_SIZE(din));
	for (; offset < last; offset += EXT2_BLOCK_SIZE(din)) {
		blkData = getBlkData(din, offset, m_inode, &blkCache, &error);
		if (error != NOERR) {
			goto END;
		}
		memcpy(blkData, writeBuf, EXT2_BLOCK_SIZE(din));
		updateBlk(0,&blkCache);
		writeBuf += EXT2_BLOCK_SIZE(din);
	}
		
	if (offset < i_offset + i_bytes) {
		// Ō̃ubNTCY
		blkData = getBlkData(din, offset, m_inode, &blkCache, &error);
		if (error != NOERR) {
			goto END;
		}
		writeBytes = i_offset + i_bytes - offset;
		memcpy(blkData, writeBuf, writeBytes);
		updateBlk(0, &blkCache);
	}
	
	error = NOERR;
END:
	freeBlkCache(din, &blkCache);
	
	return error;
}

int main(int argc, char *argv[])
{
	INODE inode;
	size_t begin;
	size_t bytes;
	uchar *writeBuf;
	int i;

	// 
	initDisk();
	initInode(&inode);

	// fBXNɏ
	writeBuf = malloc(492);
	if (writeBuf == NULL) {
		printf("malloc() error!\n");
		return 0;
	}
	for (i = 0; i < 492; ++i) {
		writeBuf[i] = i + 1;
	}
	if (writeData(0, 0, 492, writeBuf, &inode) != NOERR) {
		printf("writeData error!");
		return 0;
	}

	// FreadJnoCg readoCg
	for (;;) {
		char *readBuf;

		scanf("%u %u", &begin, &bytes);
		printf("begin=%u bytes=%u\n", begin, bytes);

		readBuf = malloc(bytes);
		if (readBuf == NULL) {
			printf("malloc error!");
			return 0;
		}
		printf("error=%d\n", readData(0, begin, bytes, readBuf, &inode));

		// fBXNj^[
		moniterDisk();
	
		// [_v
		moniterMem(readBuf, bytes);

		free(readBuf);
	}
	
	return 0;
}
