#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#define DEBUG
#define STATIC static
#define INLINE inline
#define ASSERT(bool) if ((bool) == 0){							\
	printf("Assertion! : %s line=%d \n",__FILE__, __LINE__); 	\
}
#define WARNING(bool) if ((bool) == 0){							\
	printf("Warning! : %s line=%d \n",__FILE__, __LINE__);		\
}
#define ROUNDUP(a, b) (((uint)(a) + (b) - 1) / (b) * (b))
#define ROUNDDOWN(a, b) ((uint)(a) / (b) * (b))
#define ROUNDUP_DIV(a, b) (((uint)(a) + (b) - 1) / (b))
#define PAGE_SIZE 0x1000
#define ENOMEM 5
#define printk printf

typedef unsigned char uchar;

static int enterCli()
{
	return 0;
}

static void exitCli(int a)
{
	return;
}

static void enter_spinlock(int *a)
{
	return;
}

static void exit_spinlock(int *a)
{
	return;
}

STATIC void *allocMultiKernelPage(int num)
{
	return NULL;
}

enum {
	ALLOC_TBL_MAX = 50,
	LOW_BIT = 0xffffff,
};

typedef struct {
	void *addr;
	int num;
} ALLOC_TBL;

static ALLOC_TBL allocTbl[ALLOC_TBL_MAX];
static uint highbit;

STATIC int initAlloc()
{
	void *mem = malloc(PAGE_SIZE * (ALLOC_TBL_MAX + 1));
	int i;
	
	allocTbl[0].addr = (void*) ROUNDUP((uint) mem, PAGE_SIZE);
	for (i = 1; i < ALLOC_TBL_MAX; ++i) {
		allocTbl[i].addr = (char*) allocTbl[i - 1].addr + PAGE_SIZE;
	}
	
	highbit = (uint) allocTbl[0].addr;
	highbit &= ~LOW_BIT;
}

STATIC int getAllocNum(
	void *i_addr)
{
	void *addr = (void*) ROUNDDOWN((uint) i_addr, PAGE_SIZE);
	int i;

	for (i = 0; i < ALLOC_TBL_MAX; ++i) {
		if (allocTbl[i].addr == addr) {
			return allocTbl[i].num;
		}
	}
	
	return 0;
}

STATIC void setAllocNum(
	void *i_addr,
	int num)
{
	void *addr = (void*) ROUNDDOWN((uint) i_addr, PAGE_SIZE);
	int i;

	for (i = 0; i < ALLOC_TBL_MAX; ++i) {
		if (allocTbl[i].addr == addr) {
			allocTbl[i].num = num;
			return;
		}
	}
	
	printf("setAllocNum() over!\n");
}

STATIC void *allocKernelPage()
{
	void *mem;
	int i;

	for (i = 0; i < ALLOC_TBL_MAX; ++i) {
		if (allocTbl[i].num == 0) {
			mem = allocTbl[i].addr;
			break;
		}
	}
//*****************************************************************************
	printf("allocKernelPage() 0x%x ", mem);
//*****************************************************************************
	return mem;
}

STATIC void unlinkPhysTable(void *addr)
{
	int i;

//*****************************************************************************
	printf("unlinkPhysTable() 0x%x ", addr);
//*****************************************************************************
	for (i = 0; i < ALLOC_TBL_MAX; ++i) {
		if (allocTbl[i].addr == addr) {
			allocTbl[i].num = 0;
			return;
		}
	}
}

//*******************************************************************

//================================== PRIVATE ============================================

#define HEAP_NUM() (sizeof(heapHead) / sizeof(HEAP_HEAD))

/* 
 * q[v[wb_[
 */
typedef struct MEM_HEAD{
	struct MEM_HEAD	*next;		/* ̋󂫃[擪 */
	uint			size;		// [TCY
}MEM_HEAD;

/*
 * q[vwb_[
 */
typedef struct{
	MEM_HEAD	*memTop;	/* 󂫃[擪 */
	uint		size;		/* 蓖ăTCY */
	int			gate;		/* XsbNpQ[g */
}HEAP_HEAD;

/*
 * q[vwb_[
 */
static HEAP_HEAD heapHead[]={
	{NULL, 0,      0},
	{NULL, 0x10,   0},
	{NULL, 0x20,   0},
	{NULL, 0x40,   0},
	{NULL, 0x80,   0},
	{NULL, 0x100,  0},
	{NULL, 0x200,  0},
	{NULL, 0x400,  0},
	{NULL, 0x800,  0},
	{NULL, 0x1000, 0},
};

/*
 * q[vwb_[zl
 */
static uchar heapNum1[]={
	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
	2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
	3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
	3,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
	4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
	4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
	4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
	4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
};

static uchar heapNum2[]={
	0,5,6,7,7,8,8,8,8,9,9,9,9,9,9,9,
	9,
};

/*
 * TCYq[vwb_[zl߂
 * return : zl
 */
STATIC INLINE int getHeapArray(
	size_t size)		// oCgTCY
{
	if (size < 0x100) {
		return heapNum1[size];
	}
	else {
		return heapNum2[ROUNDUP(size,0x100)>>8];
	}
}

/*
 * y[WTCYȉ̃[̎擾
 */
STATIC void *getPageMem(
	const uint i_size,
	uint *o_size)
{
	MEM_HEAD *mem;
	int eflag;
	
	// q[vwb_[zl̎擾
	int array = getHeapArray(i_size);
	
	eflag = enterCli();
	enter_spinlock(&heapHead[array].gate);
	{
		if (heapHead[array].memTop == NULL){
			MEM_HEAD *head;

			// y[W[m
			head = allocKernelPage();
			if (head == NULL){
				return NULL;
			}

			// JƂAaddress犄TCY킩悤ɂB
			setAllocNum(head, array);

			head->next = NULL;
			head->size = PAGE_SIZE;
			heapHead[array].memTop = head;
		}

		// v[̒o
		mem = heapHead[array].memTop;
		if (heapHead[array].size < mem->size) {
			MEM_HEAD *next = (MEM_HEAD*) ((uint) mem + heapHead[array].size);

			next->next = mem->next;
			next->size = mem->size - heapHead[array].size;
			heapHead[array].memTop = next;
		}
		else {
			heapHead[array].memTop = mem->next;
		}
	}
	exit_spinlock(&heapHead[array].gate);
	exitCli(eflag);

	*o_size = heapHead[array].size;

	return mem;
}

/*
 * y[WTCY傫[̎擾
 */
STATIC void *getBigMem(
	const uint i_size,
	uint *o_size)
{
	uint size = ROUNDUP(i_size, PAGE_SIZE);
	void *mem = allocMultiKernelPage(size / PAGE_SIZE);
	
	if (mem == NULL) {
		return NULL;
	}
	setAllocNum(mem, HEAP_NUM() + size / PAGE_SIZE);

	*o_size = size;

	return mem;
}

/*
 * q[v[y[WTCYȏȂ炻̕J
 * return : J̃q[v[擪|C^
 */
STATIC MEM_HEAD *freePageSize(
	MEM_HEAD *m_src)
{
	uint srcAddr = (uint) m_src;
	uint pageBoundAddr = ROUNDUP(srcAddr, PAGE_SIZE);	// y[WEAhX

	if (m_src == NULL) {
		return NULL;
	}

	if (PAGE_SIZE <= (int) m_src->size - (int) (pageBoundAddr - srcAddr)) {
		MEM_HEAD *top;

/****************************************************************************************************************/
fprintf(stderr,"freePageSize() src=%x size=%x ", m_src, m_src->size);
/****************************************************************************************************************/
		if (pageBoundAddr + PAGE_SIZE < srcAddr + m_src->size) {
			MEM_HEAD *next = (MEM_HEAD*) (pageBoundAddr + PAGE_SIZE);
			
/****************************************************************************************************************/
fprintf(stderr,"next=%x ", next);
/****************************************************************************************************************/
			next->next = m_src->next;
			next->size = srcAddr + m_src->size - pageBoundAddr - PAGE_SIZE;
			m_src->next = next;
		}

		if (srcAddr == pageBoundAddr) {
			top = m_src->next;
		}
		else {
			m_src->size = pageBoundAddr - srcAddr;
			top = m_src;
		}

		unlinkPhysTable((void*) pageBoundAddr);
		
/****************************************************************************************************************/
fprintf(stderr,"\n");
/****************************************************************************************************************/
		return top;
	}
	else {
		return m_src;
	}
}

/*
 * y[WTCYȉ̃[J
 */
STATIC void freePageMem(
	const int array,
	MEM_HEAD *i_src)
{
	int eflag;
	
	eflag = enterCli();
	enter_spinlock(&heapHead[array].gate);
	{
		MEM_HEAD *src = i_src;
		MEM_HEAD *next;
		MEM_HEAD *prev;

/****************************************************************************************************************/
fprintf(stderr,"freePageMem() ");
/****************************************************************************************************************/
		/*
		 * q[v[ɐڑ
		 * out : prev = Õq[vANULLȂ玟̃q[v͐擪 src = ̃q[v
		 */
		src->size = heapHead[array].size;
		prev = NULL;
		for (next = heapHead[array].memTop;; next = next->next) {
/****************************************************************************************************************/
fprintf(stderr,"next=%x ", next);
/****************************************************************************************************************/
		// AhX͂肦Ȃ
			ASSERT(src != next);

			if (next == NULL) {
				src->next = NULL;
				break;
			}

			if ((uint) next + next->size == (uint) src) {
/****************************************************************************************************************/
fprintf(stderr,"1 ");
/****************************************************************************************************************/
				// Õq[vƌ
				next->size += src->size;
				
				src = next;
			}
			else {
				if (src < next) {
/****************************************************************************************************************/
fprintf(stderr,"2 ");
/****************************************************************************************************************/
					if ((uint) src + src->size == (uint) next) {
/****************************************************************************************************************/
fprintf(stderr,"3 ");
/****************************************************************************************************************/
						// ̃q[vƌ
						src->size += next->size;
						src->next = next->next;
					}
					else {
/****************************************************************************************************************/
fprintf(stderr,"4 ");
/****************************************************************************************************************/
						src->next = next;
					}
					
					break;
				}

				prev = next;
			}
		}

		// q[v[y[WTCYȏȂ炻̕J
		src = freePageSize(src);
		if (prev == NULL) {
			heapHead[array].memTop = src;
		}
		else {
			prev->next = src;
		}
	}
	exit_spinlock(&heapHead[array].gate);
	exitCli(eflag);
/****************************************************************************************************************/
fprintf(stderr,"\n");
/****************************************************************************************************************/
}

//================================== PUBLIC =============================================

/*
 * vTCY[蓖āATCYԂ
 * return : allocate memory or NULL
 */
void *kmallocSize(
	const uint i_size,
	uint *o_size)
{
	ASSERT(o_size != NULL);

	if (i_size == 0){
		*o_size = 0;
		return NULL;
	}

	if (i_size <= PAGE_SIZE) {
		return getPageMem(i_size, o_size);
	}
	else {
		return getBigMem(i_size, o_size);
	}
}

/*
 * vTCY[蓖Ă
 * return : allocate address or NULL
 */
void *kmalloc(size_t size)
{
	int realSize;

	return kmallocSize(size, &realSize);
}

/*
 * [J
 * return : J[TCY
 */
uint kfree(void *ptr)
{
	int array;

/****************************************************************************************************************/
fprintf(stderr,"kfree() ");
/****************************************************************************************************************/
	if (ptr == NULL){
		return 0;
	}

	array = getAllocNum(ptr);

	if (array == 0){
/****************************************************************************************************************/
fprintf(stderr,"1\n");
/****************************************************************************************************************/
#ifdef DEBUG
		printk("kfree() 0x%x did not allocated : call function=0x%x\n",ptr, *((uint*)&ptr - 1));
#endif
		return 0;
	}

	if (array < HEAP_NUM()) {
/****************************************************************************************************************/
fprintf(stderr,"2 ");
/****************************************************************************************************************/
		/*
		 * JTCYy[WTCYȉ
		 */
#ifdef DEBUG
		// J|C^TCYEłȂ
		if (0 < ((uint) ptr % heapHead[array].size)) {
			printk("kfree() 0x%x is wrong addres! : call function=0x%x\n",ptr, *((uint*)&ptr - 1));
			return 0;
		}
//**************************************************************************************
//test_mbuf((uint)ptr, *((uint*) &ptr - 1));
//**************************************************************************************
		// QƂƃy[WtHgNlŖ߂Ă
		memset(ptr, 0xf5, heapHead[array].size);
#endif
		freePageMem(array, ptr);

		return heapHead[array].size;
	}
	else {
		uint size;
		uint mem;
		int i;
/****************************************************************************************************************/
fprintf(stderr,"3 ");
/****************************************************************************************************************/
#ifdef DEBUG
		// J|C^TCYEłȂ
		if (0 < ((uint) ptr % PAGE_SIZE)) {
			printk("kfree() 0x%x is wrong addres! : call function=0x%x\n",ptr, *((uint*)&ptr - 1));
			return 0;
		}
#endif
		size = (array - HEAP_NUM()) * PAGE_SIZE;
		mem = (uint) ptr;

		for (i = array - HEAP_NUM(); 0 < i ; --i){
			unlinkPhysTable((void*) mem);
			mem += PAGE_SIZE;
		}
		
		return size;
	}
}

//*************************************************************************************************

static void dump()
{
	int i;
	
	for (i = 1; i < HEAP_NUM(); ++i) {
		MEM_HEAD *next;
	
		printf("size:0x%x ", heapHead[i].size);
		for (next = heapHead[i].memTop; next != NULL; next = next->next) {
			printf("0x%x-0x%x ", (uint) next, next->size);
		}
		printf("\n");
	}
	printf("\n");
}

int main(int argc, char *argv[])
{
	initAlloc();
	
	for (;;) {
		int func;
		uint size_addr;
		void *mem;

		scanf("%d %x", &func, &size_addr);

		switch (func) {
		case 1:
			mem = kmalloc(size_addr);
			printf("allocate %x\n", (uint) mem);
			dump();
			break;
		case 2:
			size_addr |= highbit;
			kfree((void*) size_addr);
			dump();
			break;
		default:
			break;
		}
	}
	
	return 0;
}
