#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "AggregateList.h"
#include "IteratorList.h"

#include "test.h"


/******************************* eXgp` ******************************************/

enum {
	PCATCH = 0x100,
	REAL_TIMER_RATE	= 500,
};

typedef struct {
	AggregateList	aggregate;	// ^C}[NW
	int				count;		// N^C}[
	int				lock;		// XsbNQ[g
} TIMER_OWNER;

struct callout_handle {
	void *callout;
};

typedef void timeout_t (void *);

typedef struct {
	TIMER_OWNER procTimer;
} PROC;
static PROC proc;

struct buf {
	void *proc;
};

static PROC *get_current_task()
{
	return &proc;
}

static void sendSignal(void *dummy1, int dummy2)
{
	printf("sendSignal()\n"); 
}

static void forceSendSignal(void *dummy1, int dummy2, int dummy3)
{
	printf("forceSendSignal()\n"); 
}

static void setIntervalTasking(int dummy)
{
	printf("setIntervalTasking()\n"); 
}

static void sigWait() { return; }

static uint getIntervalTimerPastTime()
{
	return 0;
}

static uint getRealTimerPastTime()
{
	return 0;
}

static void setIntervalTimer(const uint dummy)
{
	return;
}

static void setRealTimer(const uint dummy)
{
	return;
}

static void timeoutftn(void *dummy)
{
	sendSignal(NULL, 0);
}

static void displayTop() {return;}

static int isSigint() {return 0;}

static void dump();

/******************************* eXgR[h ******************************************/

//================================== ʒ` ===========================================

/*
 * ^C}[wb_
 */
typedef struct {
	AggregateList	aggregate;					// ^C}[NW
	uint			(*getPastTime)();			// ^C}[荞݂܂ł݂̌̌oߎms擾
	void			(*setTimer)(const uint);	// ^C}[ms̐ݒ
	int				lock;						// XsbNQ[g
} TIMER_HEAD;

/*
 * ^C}[\
 */
typedef struct {
	uint		init;				// Đڑ^C}[l
	void		(*handle)(void*);	// nh
	void		*arg;				// nh
	List		ownedList;			// LNXg

	// ^C}[Ǘ
	uint		count;				// c莞
	List		timerList;			// ^C}[Xg
	TIMER_HEAD	*timerHead;			// ڑĂ^C}[wb_AڑĂȂNULL
} TIMER;

/***************************************************************************************************
 * ^C}[wb_
 ***************************************************************************************************/

//================================== PRIVATE =============================================

/*
 * C^[o^C}[Nwbh
 */
static TIMER_HEAD intervalTimer;

/*
 * A^C}[Nwbh
 */
static TIMER_HEAD realTimer;

/*
 * ^C}[wb_̏
 */
STATIC void initTimerHead()
{
	// C^[o^C}[wb_̏
	AggregateListConstructor(&intervalTimer.aggregate);
	intervalTimer.getPastTime = getIntervalTimerPastTime;
	intervalTimer.setTimer = setIntervalTimer;
	intervalTimer.lock = 0;
	
	// A^C}[wb_̏
	AggregateListConstructor(&realTimer.aggregate);
	realTimer.getPastTime = getRealTimerPastTime;
	realTimer.setTimer = setRealTimer;
	realTimer.lock = 0;
}

/*
 *sXt
 * ^C}[\̂ԏɐڑB
 */
STATIC void linkTimer(
	const int i_time,			// ^C}[ms
	TIMER *m_timer,				// ڑ^C}[\
	TIMER_HEAD *m_timerHead)	// ڑ^C}[wb_
{
	uint time = i_time;
	AggregateList *aggregate = &m_timerHead->aggregate;
	IteratorList *iterator;

	iterator = aggregate->iterator(aggregate);
	for (;;) {
		TIMER *timer;
		
		if (iterator->hasNext(iterator) == BOOL_FALSE) {
			aggregate->insertEnd(aggregate, &m_timer->timerList);
			break;
		}

		timer = iterator->next(iterator);
		if (time <= timer->count) {
			timer->count -= time;
			aggregate->insertPrev(aggregate, &timer->timerList, &m_timer->timerList);
			break;
		}
		else {
			time -= timer->count;
		}
	}
	m_timer->count = time;
}

/*
 *sXt
 * ^C}[\̂wb_N؂藣B
 */
STATIC void unlinkTimer(
	TIMER *m_timer,				// ؂藣^C}[\
	TIMER_HEAD *m_timerHead)	// ڑ^C}[wb_
{
	AggregateList *aggregate = &m_timerHead->aggregate;
	TIMER *next = aggregate->refNext(aggregate, &m_timer->timerList);

	ASSERT(m_timer != NULL);

	if (next != NULL) {
		next->count += m_timer->count;
	}
	aggregate->removeEntry(aggregate, &m_timer->timerList);
}

/*
 *stsXt
 * Ԃo߂^C}[nhsB
 */
STATIC void execTimer(
	const uint i_time,						// oߎms
	TIMER_HEAD *m_timerHead,				// ڑ^C}[wb_
	AggregateList *m_aggregateInitList,		// ĐڑKvȃ^C}[\̂ڑPXgwb_
	AggregateList *m_aggregateUnlinkList)	// 폜^C}[\̂ڑPXgwb_
{
	uint time = i_time;
	AggregateList *aggregate = &m_timerHead->aggregate;
	TIMER *timer;

	for (timer = aggregate->refHead(aggregate); timer != NULL;) {
		TIMER *next = aggregate->refNext(aggregate, &timer->timerList);

		if (timer->count <= time) {
			aggregate->removeEntry(aggregate, &timer->timerList);
			time -= timer->count;
			timer->count = 0;

			// ^C}[nh̎s
			timer->handle(timer->arg);

			// l΍Đڑ
			if (0 < timer->init) {
				m_aggregateInitList->insertHead(m_aggregateInitList, &timer->timerList);
			}
			else {
				m_aggregateUnlinkList->insertHead(m_aggregateUnlinkList, &timer->timerList);
			}
		}
		else {
			timer->count -= time;
			break;
		}
		
		timer = next;
	}
}

/*
 *sXt
 * ^C}[N̐ݒ莞Ԃ炷
 */
STATIC void subtractTimeFromTimerLink(
	const uint time,			// ms
	TIMER_HEAD *i_timerHead)	// ڑ^C}[wb_
{
	AggregateList *aggregate = &i_timerHead->aggregate;
	TIMER *timer = aggregate->refHead(aggregate);
	
	ASSERT(timer != NULL);
	ASSERT(time <= timer->count);
	
	timer->count -= time;
}

/*
 *sQt
 * ^C}[N̍ĒZԂ擾
 * return : ĒZms
 */
STATIC uint getMinTimeFromTimerLink(
	TIMER_HEAD *i_timerHead)	// ڑ^C}[wb_
{
	AggregateList *aggregate = &i_timerHead->aggregate;
	TIMER *timer = aggregate->refHead(aggregate);
	
	ASSERT(timer != NULL);
	
	return timer->count;
}

/*
 * ^C}[No^tOݒ肷
 */
STATIC INLINE void setRegist(
	TIMER *m_timer,
	TIMER_HEAD *timerHead)
{
	m_timer->timerHead = timerHead;
}

/*
 * ^C}[No^tO
 */
STATIC INLINE void resetRegist(
	TIMER *m_timer)
{
	m_timer->timerHead = NULL;
}

/*
 * ^C}[ݒ肷
 */
STATIC void setTimerLink(
	const uint time,			// ݒ莞
	TIMER *m_timer,				// ^C}[\
	TIMER_HEAD *m_timerHead)	// ڑ^C}[wb_
{
	// ^C}[s܂ł݂̌̌oߎԂ擾
	uint pastTime = m_timerHead->getPastTime();
	
	// ^C}[Nɓo^
	linkTimer(time + pastTime, m_timer, m_timerHead);
	setRegist(m_timer, m_timerHead);

	// ^C}[ĒZԂɍĐݒ
	if (time + pastTime == getMinTimeFromTimerLink(m_timerHead)) {
		m_timerHead->setTimer(time);
		subtractTimeFromTimerLink(pastTime, m_timerHead);
	}
}

/*
 *stsXt
 * Ԃo߂^C}[nhsB
 */
STATIC void doTimer(
	const uint i_time,			// oߎ
	TIMER_HEAD *m_timerHead)	// ڑwb_
{
	AggregateList aggregateInitTimer;		// PXg
	AggregateList aggregateUnlilnkTimer;	// PXg
	TIMER *timer;
	
	AggregateSingleListConstructor(&aggregateInitTimer);
	AggregateSingleListConstructor(&aggregateUnlilnkTimer);

	// ^C}[̎s
	execTimer(i_time, m_timerHead, &aggregateInitTimer, &aggregateUnlilnkTimer);

	// Đڑ^C}[ڑ
	while ((timer = aggregateInitTimer.getHead(&aggregateInitTimer)) != NULL) {
		setTimerLink(timer->init, timer, m_timerHead);
	}
	
	while ((timer = aggregateUnlilnkTimer.getHead(&aggregateUnlilnkTimer)) != NULL) {
		resetRegist(timer);
	}
}

//================================== PROTECT =============================================

/*
 *sQt
 * ^C}[NɐڑĂ邩H
 * return : YES or NO
 */
STATIC INLINE int isRegistTimer(
	const TIMER *timer)
{
	return (timer->timerHead != NULL) ? YES : NO;
}

/*
 *sӁt̊֐͕K delFromSchedule() ̌ŎgƁBAɃ^CAEgNƃvZXAȂ댯B
 *sXt
 * ^C}[\̂^C}[wb_ɓo^
 * return : error number
 */
STATIC void registTimer(
	const uint i_time,	// ^C}[ݒ莞ms
	TIMER  *m_timer)
{
	TIMER_HEAD *timerHead;
	int eflags;

	// ^C}[wb_̐ݒ
	if (i_time < REAL_TIMER_RATE) {
		timerHead = &intervalTimer;
	}
	else {
		timerHead = &realTimer;
	}

	eflags = enterCli();
	enter_spinlock(&timerHead->lock);
	{
		setTimerLink(i_time, m_timer, timerHead);
	}
	exit_spinlock(&timerHead->lock);
	exitCli(eflags);
}

/*
 * ^C}[\̂^C}[wb_폜
 */
STATIC void unregistTimer(
	TIMER  *m_timer)
{
	TIMER_HEAD *timerHead = m_timer->timerHead;
	int eflags;

	ASSERT(m_timer != NULL);
	
	if (isRegistTimer(m_timer) == YES) {
		eflags = enterCli();
		enter_spinlock(&timerHead->lock);
		{
			if (isRegistTimer(m_timer) == YES) {
				unlinkTimer(m_timer, m_timer->timerHead);
				resetRegist(m_timer);
			}
		}
		exit_spinlock(&timerHead->lock);
		exitCli(eflags);
	}
}

/*
 *sQt
 * c莞Ԃ擾
 * return : c莞ms
 */
STATIC uint getRemainTime(
	TIMER *i_timer)
{
	TIMER_HEAD *timerHead = i_timer->timerHead;
	uint time;
	int eflags;
	
	ASSERT(i_timer != NULL);

	time = 0;
	if (isRegistTimer(i_timer) == YES) {
		eflags = enterCli();
		enter_spinlock(&timerHead->lock);
		{
			if (isRegistTimer(i_timer) == YES) {
				AggregateList *aggregate = &timerHead->aggregate;
				IteratorList *iterator;

				iterator = aggregate->iterator(aggregate);
				for (;;) {
					TIMER *timer = iterator->next(iterator);
					
					// i_timer݂͂
					ASSERT(timer != NULL);
					
					// c莞ԂZ
					time += timer->count;
					
					if (timer == i_timer) {
						break;
					}
				}
			}
		}
		exit_spinlock(&timerHead->lock);
		exitCli(eflags);
	}

	return time;
}

//================================== PUBLIC =============================================

/*
 *st
 * C^[o^C}[̎s
 */
void doIntervalTimer(
	const uint time)		// o߂(ms)
{
	enter_spinlock(&intervalTimer.lock);
	{
		doTimer(time, &intervalTimer);
	}
	exit_spinlock(&intervalTimer.lock);
}

/*
 *st
 * ANbN^C}[̎s
 */
void doRealTimer(
	const uint time)		// o߂(ms)
{
	enter_spinlock(&realTimer.lock);
	{
		doTimer(time, &realTimer);
	}
	exit_spinlock(&realTimer.lock);
}

/***************************************************************************************************
 * ^C}[
 *
 * PXg
 ***************************************************************************************************/

//================================== PRIVATE ============================================

enum{
	TIMER_MAX = 10,			// ^C}[ő吔
};

static TIMER_OWNER systemTimer;	// VXeێ^C}[

/*
 * ^C}[Lwb_̏
 */
STATIC INLINE void initOwnerHead(
	TIMER_OWNER *m_owner)
{
	AggregateSingleListConstructor(&m_owner->aggregate);
	m_owner->count = 0;
	m_owner->lock = 0;
}

/*
 *sQt
 * Lwb_nhŒTB
 * return : NOERR or -ENOENT
 */
STATIC int findTimerByHandle(
	TIMER_OWNER *i_owner,		// Lwb_
	void (*handle)(void*),
	TIMER **o_timer)
{
	AggregateList *aggregate = &i_owner->aggregate;
	IteratorList *iterator;

	iterator = aggregate->iterator(aggregate);
	while (iterator->hasNext(iterator) == BOOL_TRUE) {
		TIMER *timer = iterator->next(iterator);
		
		if (timer->handle == handle) {
			*o_timer = timer;
			return NOERR;
		}
	}

	return -ENOENT;
}

/*
 * Lwb_폜σ^C}[T
 * return : NOERR or -ENOENT
 */
STATIC int findDelTimer(
	TIMER_OWNER *i_owner,		// Lwb_
	TIMER **o_timer)
{
	AggregateList *aggregate = &i_owner->aggregate;
	IteratorList *iterator;

	iterator = aggregate->iterator(aggregate);
	while (iterator->hasNext(iterator) == BOOL_TRUE) {
		TIMER *timer = iterator->next(iterator);
		
		if (isRegistTimer(timer) == NO) {
			*o_timer = timer;
			return NOERR;
		}
	}

	return -ENOENT;
}

/*
 * ^C}[蓖ď
 * return : error number
 */
STATIC int allocTimer(
	const uint initTime,	// ԌoߌɍĂѐݒ肷^C}[
	void (*handle)(void*),	// ԌoߎɎs鑀
	void *arg,				// ԌoߎɎs鑀̈
	TIMER_OWNER *m_owner,	// ^C}[L
	TIMER **o_timer)
{
	TIMER *timer;

	ASSERT(handle != NULL);

	// 󂫃^C}[TĂȂΊ蓖Ă
	if (findDelTimer(m_owner, &timer) == -ENOENT) {
		AggregateList *aggregate = &m_owner->aggregate;

		if (TIMER_MAX <= m_owner->count) {
			WARNING(0);
			return -ENOMEM;
		}

		timer = kmalloc(sizeof(*timer));
		if (timer == NULL){
			return -ENOMEM;
		}
		listConstructor(&timer->ownedList, timer);
		aggregate->insertHead(aggregate, &timer->ownedList);
	}

	listConstructor(&timer->timerList, timer);
	timer->init = initTime;
	timer->handle = handle;
	timer->arg = arg;
	++m_owner->count;

	*o_timer = timer;

	return NOERR;
}

//================================== PROTECT ============================================

/*
 * vZX^C}[̏
 */
STATIC void initProcessTimer(PROC *proc)
{
	initOwnerHead(&proc->procTimer);
}

/*
 * VXe^C}[蓖Ă
 * return : error number
 */
STATIC int allocSystemTimer(
	const uint i_time,		// ^C}[
	const uint initTime,	// ԌoߌɍĂѐݒ肷^C}[
	void (*handle)(void*),	// ԌoߎɎs鑀
	void *arg,				// ԌoߎɎs鑀̈
	TIMER **o_timer)
{
	int error;

	enter_spinlock(&systemTimer.lock);
	{
		error = allocTimer(initTime, handle, arg, &systemTimer, o_timer);
	}
	exit_spinlock(&systemTimer.lock);
	
	if (error == NOERR) {
		registTimer(i_time, *o_timer);
		return NOERR;
	}
	else {
		return error;
	}
}

/*
 *sXt
 * vZX^C}[蓖Ă
 * return : error number
 */
STATIC int allocProcessTimer(
	const uint i_time,		// ^C}[
	const uint initTime,	// ԌoߌɍĂѐݒ肷^C}[
	void (*handle)(void*),	// ԌoߎɎs鑀
	void *arg,				// ԌoߎɎs鑀̈
	TIMER **o_timer)
{
	int error = allocTimer(initTime, handle, arg, &get_current_task()->procTimer, o_timer);

	if (error == NOERR) {
		registTimer(i_time, *o_timer);
		return NOERR;
	}
	else {
		return error;
	}
}

/*
 * ^C}[J
 * return : error number
 */
STATIC void freeTimer(
	TIMER *m_timer)
{
	// o^
	unregistTimer(m_timer);
	
	// ^C}[\̂͊JȂ
}

/*
 *sXt
 * vZXLS^C}[J
 */
STATIC void releaseProcessTimer(
	TIMER_OWNER *m_owner)
{
	AggregateList *aggregate = &m_owner->aggregate;
	TIMER *timer;

	while ((timer = aggregate->getHead(aggregate)) != NULL) {
		// o^
		unregistTimer(timer);

		kfree(timer);
	}
}

/*
 *sQt
 * A[nh
 * return : ^C}[c莞ms
 */
STATIC INLINE uint getRemainTimeFromTimer(
	TIMER *i_timer)
{
	return getRemainTime(i_timer);
}

/*
 *st
 * A[nh
 */
STATIC void handleAlarm(void *proc)
{
	sendSignal(proc, SIGALRM);
}

/*
 *st
 * ^XNNnh
 */
STATIC void handleWake(void *proc)
{
	forceSendSignal(proc, SIGWAKE, TASK_SIGNAL_WAIT);
}

/*
 * ^XNNnh
 */
STATIC void handleDevWake(void *proc)
{
//	forceSendSignal(proc, SIGWAKE, TASK_SIGNAL_WAIT | TASK_DEVICE_WAIT);
	forceSendSignal(proc, SIGWAKE, TASK_DEVICE_WAIT);
}

/*
 *st
 * ^XLOnh
 */
STATIC void handleTasking(
	void *dummy)		// gp
{
	// ^XNXCb`s
	setIntervalTasking(YES);
/********************************************************************/
	{
		static char value = '0';

		if ('9' < value){
			value = '0';
		}
		displayTop(60, &value, 1);
		++value;
	}
/********************************************************************/
}

//================================== PUBLIC ==============================================

/*
 * foCX荞ݑ҂N^C}[̃Zbg
 * return : error number
 */
int setDevWakeTimer(
	const uint i_time,		// ^C}[ԁimsj
	void **o_timer)
{
	ASSERT(0 < i_time);

	// ^C}[̍쐬
	return allocProcessTimer(i_time, 0, handleDevWake, get_current_task(), (TIMER**) o_timer);
}

/*
 * N^C}[̉
 */
void delWakeTimer(
	void *i_timer)		// ^C}[
{
	freeTimer(i_timer);
}

/*
 * ^XLO̐ݒ
 * return : error number
 */
int setTasking(
	const uint i_time)		// ^XLOԁimsj
{
	TIMER *timer;
	
	ASSERT(0 < i_time);

	// ^C}[̍쐬
	return allocSystemTimer(i_time, i_time, handleTasking, NULL, &timer);
}

/*
 * timeout --
 *	Execute a function after a specified length of time.
 *
 * untimeout --
 *	Cancel previous timeout function call.
 *
 * callout_handle_init --
 *	Initialize a handle so that using it with untimeout is benign.
 *
 *	See AT&T BCI Driver Reference Manual for specification.  This
 *	implementation differs from that one in that although an 
 *	identification value is returned from timeout, the original
 *	arguments to timeout as well as the identifier are used to
 *	identify entries for untimeout.
 *
 *st
 *sӁtvZXĂ΂ꍇ
 */
void callout_handle_init(struct callout_handle *callout)
{
	callout->callout = NULL;
}

struct callout_handle timeout(
	timeout_t *ftn,
	void *arg, 
	int to_ticks)		// ms
{
	TIMER *timer;
	struct callout_handle callout;

	if (to_ticks == 0) {
		callout.callout = NULL;
		return callout;
	}

	if (allocSystemTimer(to_ticks, 0, ftn, arg, &timer) == NOERR) {
		callout.callout = timer;
	}
	else {
		printk("Failed timeout()!\n");
		callout.callout = NULL;
	}

	return callout;
}

/*
 *st
 */
void untimeout(
	timeout_t *ftn, 
	void *arg, 
	struct callout_handle callout)
{
	if (callout.callout == NULL){
		return;
	}

	freeTimer(callout.callout);
}

void callout_init(void *c)
{
	// Ƃ肠Ȃ
}

void callout_reset(void *c, int to_ticks, void (*ftn)(void *), void *arg)
{
	// Ƃ肠Ȃ
}

/*
 * return : 0 : Xgbvcallout or 1 : XgbvcalloutL
 */
int callout_stop(void *c)
{
	// Ƃ肠Ȃ
	return 0;
}

/*
 *sFreeBSDt
 * General sleep call.  Suspends the current process until a wakeup is
 * performed on the specified identifier.  The process will then be made
 * runnable with the specified priority.  Sleeps at most timo/hz seconds
 * (0 means no timeout).  If pri includes PCATCH flag, signals are checked
 * before and after sleeping, else signals are not checked.  Returns 0 if
 * awakened, EWOULDBLOCK if the timeout expires.  If PCATCH is set and a
 * signal needs to be delivered, ERESTART is returned if the current system
 * call should be restarted if possible, and EINTR is returned if the system
 * call should be interrupted by the signal (return EINTR).
 */
int	tsleep(
	void *chan, 
	int pri, 
	const char *wmesg, 
	int timeout)
{
	struct buf *bp = chan;
	TIMER *timer;
	uint remainTime;
	int error;

	bp->proc = get_current_task();

	if (0 < timeout) {
		error = allocProcessTimer(timeout, 0, handleWake, get_current_task(), &timer);
		if (error != NOERR) {
			return 0;
		}
	}
	else if (timeout == 0) {
		// ^CAEgȂ
		error = allocProcessTimer(UINT_MAX, UINT_MAX, handleWake, get_current_task(), &timer);
		if (error != NOERR){
			return 0;
		}
	}
	else {
		return 0;
	}
	sigWait();
	if ((pri & PCATCH) != 0) {
		if (isSigint() == TRUE) {
			return EINTR;
		}
	}

	// c莞Ԃ擾
	remainTime = getRemainTimeFromTimer(timer);

	// 폜
	freeTimer(timer);
	
	if (remainTime == 0) {
		return EWOULDBLOCK;
	}
	else {
		return 0;
	}
}

int sys_sleep(uint seconds)
{
	TIMER *timer;
	uint remainTime;
	int error;

	if (seconds == 0){
		return 0;
	}

	error = allocProcessTimer(seconds * 1000, 0, handleWake, get_current_task(), &timer);
	if (error != NOERR){
		return seconds;
	}

	sigWait();

	// c莞Ԃ擾
	remainTime = getRemainTimeFromTimer(timer);

	// 폜
	freeTimer(timer);

	return remainTime;
}

int sys_alarm(uint seconds)
{
	TIMER *timer;

	// A[^C}[ɂΉ
	if (findTimerByHandle(&get_current_task()->procTimer, handleAlarm, &timer) == NOERR) {
		freeTimer(timer);
	}

	// vZX^C}[ݒ
	if (0 < seconds){
		return allocProcessTimer(seconds*1000, 0, handleAlarm, get_current_task(), &timer);
	}
	else {
		return 0;
	}
}

/******************************* _v ******************************************/

static char *getTimerName(
	TIMER_HEAD *timerHead)
{
	if (timerHead == &intervalTimer) {
		return "intervalTimer";
	}
	else if (timerHead == &realTimer) {
		return "realTimer";
	}
	else {
		return "Unknown";
	}
}

static char *getHandleName(
	void handle(void*))
{
	if (handle == handleAlarm) {
		return "handleAlarm";
	}
	else if (handle == handleWake) {
		return "handleWake";
	}
	else if (handle == handleDevWake) {
		return "handleDevWake";
	}
	else if (handle == handleTasking) {
		return "handleTasking";
	}
	else if (handle == timeoutftn) {
		return "timeoutftn";
	}
	else {
		return "Unknown";
	}
}

static void printEntry(
	TIMER *timer)
{
	printf("    0x%06x-%04d-%04d-%s-%s\n",
		timer,
		timer->count,
		timer->init,
		getHandleName(timer->handle),
		getTimerName(timer->timerHead));
}

static void dumpList(
	AggregateList *aggregate)
{
	if (aggregate->refHead(aggregate) != NULL) {
		IteratorList *iterator = aggregate->iterator(aggregate);
		while (iterator->hasNext(iterator) == BOOL_TRUE) {
			TIMER *timer = iterator->next(iterator);
			printEntry(timer);
		}
	}
	else {
		printf("    No List\n");
	}
}

static void dump()
{
	printf("Interval timer\n");
	dumpList(&intervalTimer.aggregate);

	printf("Real timer\n");
	dumpList(&realTimer.aggregate);

	printf("System timer\n");
	dumpList(&systemTimer.aggregate);

	printf("Process timer\n");
	dumpList(&proc.procTimer.aggregate);

	printf("\n");
}

static TIMER *getTimer(
	const int index,	// Xg̏ԂP`
	AggregateList *aggregate)
{
	IteratorList *iterator = aggregate->iterator(aggregate);
	int i = 0;
	
	while (iterator->hasNext(iterator) == BOOL_TRUE) {
		TIMER *timer = iterator->next(iterator);
		if (index == ++i) {
			return timer;
		}
	}

	return NULL;
}

/******************************************************************************************/

int main(int argc, char *argv[])
{
	// VXe^C}[Ǘwb_̏
	initTimerHead();

	// VXe^C}[Lwb_̏
	initOwnerHead(&systemTimer);

	// vZX^C}[Lwb_̏
	initProcessTimer(&proc);

	for (;;) {
		int func;
		uint time;		// msP

		scanf("%d %d", &func, &time);

		switch (func) {
		case 11: {
			void *timer;
			setDevWakeTimer(time, &timer);
			dump();
			break;
		}
		case 12: {
			// time̓^C}[LXg̏ԍɓǂݑウ
			int index = time;
			TIMER *timer;
			timer = getTimer(index, &proc.procTimer.aggregate);
			if (timer != NULL) {
				delWakeTimer(timer);
				dump();
			}
			else {
				printf("Index %d is out of range!\n", index);
			}
			break;
		}
		case 13: {
			struct buf buf;
			tsleep(&buf, 0, NULL, time);
			dump();
			break;
		}
		case 14:
			sys_sleep(time);	// time = seconds
			dump();
			break;
		case 15:
			sys_alarm(time);	// time = seconds
			dump();
			break;
		case 16:
			releaseProcessTimer(&get_current_task()->procTimer);
			dump();
			break;
		case 21:
			setTasking(time);
			dump();
			break;
		case 22:
			timeout(timeoutftn, NULL, time);
			dump();
			break;
		case 23: {
			// time̓^C}[LXg̏ԍɓǂݑウ
			int index = time;
			struct callout_handle callout;
			callout.callout = getTimer(index, &systemTimer.aggregate);
			if (callout.callout != NULL) {
				untimeout(NULL, NULL, callout);
				dump();
			}
			else {
				printf("Index %d is out of range!\n", index);
			}
			break;
		}
		case 31:
			doIntervalTimer(time);
			dump();
			break;
		case 32:
			doRealTimer(time);
			dump();
			break;
		default:
			printf("Unknown func number : %d!\n", func);
			break;
		}
	}

	return 0;
}
