/*
 * timer.c
 *
 * Copyright 2007, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 *
 */


#include <sys/config.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/buf.h>
#include <sys/callout.h>
#include <sys/systm.h>
#include <sys/limits.h>
#include <machine/lib.h>
#include <dev/8254.h>
#include <dev/MC146818.h>
#include <kern/lock.h>
#include <kern/proc.h>
#include <kern/signal.h>
#include <kern/mm.h>
#include <kern/time.h>
#include <kern/AggregateList.h>
#include <kern/timer.h>

#include <kern/debug.h>


//#define DEBUG_TIMER 1
#ifdef DEBUG_TIMER
	#define STATIC
	#define INLINE
#else
	#define STATIC	static
	#define INLINE	inline
#endif


//==================================  ===========================================

/*
 * ޡإå
 */
typedef struct {
	AggregateList	aggregate;					// ޡ󥯽
	uint			(*getPastTime)();			// ޡߤޤǤθߤηвms
	void			(*setTimer)(const uint);	// ޡms
	int				lock;						// ԥå
} TIMER_HEAD;

/*
 * ޡ¤
 */
typedef struct {
	uint		init;				// ³ޡ
	void		(*handle)(void*);	// ϥɥ
	void		*arg;				// ϥɥ
	List		ownedList;			// ͭ󥯥ꥹ

	// ޡ
	uint		count;				// Ĥ
	List		timerList;			// ޡꥹ
	TIMER_HEAD	*timerHead;			// ³Ƥ륿ޡإå³ƤʤNULL
} TIMER;

/***************************************************************************************************
 * ޡ
 ***************************************************************************************************/

//================================== PRIVATE =============================================

/*
 * ֤
 * return : ms
 */
STATIC uint getInitTime(
	const TIMER *i_timer)
{
	return i_timer->init;
}

/***************************************************************************************************
 * ޡإå
 ***************************************************************************************************/

//================================== PRIVATE =============================================

/*
 * 󥿡Х륿ޡ󥯥إå
 */
static TIMER_HEAD intervalTimer;

/*
 * ꥢ륿ޡ󥯥إå
 */
static TIMER_HEAD realTimer;

/*
 * ޡإåν
 */
STATIC void initTimerHead()
{
	// 󥿡Х륿ޡإåν
	AggregateListConstructor(&intervalTimer.aggregate);
	intervalTimer.getPastTime = getIntervalTimerPastTime;
	intervalTimer.setTimer = setIntervalTimer;
	intervalTimer.lock = 0;
	
	// ꥢ륿ޡإåν
	AggregateListConstructor(&realTimer.aggregate);
	realTimer.getPastTime = getRealTimerPastTime;
	realTimer.setTimer = setRealTimer;
	realTimer.lock = 0;
}

/*
 *Թ
 * ޡ¤Τֽ³롣
 */
STATIC void linkTimer(
	const int i_time,			// ޡms
	TIMER *m_timer,				// ³륿ޡ¤
	TIMER_HEAD *m_timerHead)	// ³西ޡإå
{
	uint time = i_time;
	AggregateList *aggregate = &m_timerHead->aggregate;
	IteratorList iterator;

	aggregate->iterator(aggregate, &iterator);
	for (;;) {
		TIMER *timer;
		
		if (iterator.hasNext(&iterator) == BOOL_FALSE) {
			aggregate->insertEnd(aggregate, &m_timer->timerList);
			break;
		}

		timer = iterator.next(&iterator);
		if (time <= timer->count) {
			timer->count -= time;
			aggregate->insertPrev(aggregate, &timer->timerList, &m_timer->timerList);
			break;
		}
		else {
			time -= timer->count;
		}
	}
	m_timer->count = time;
}

/*
 *Թ
 * ޡ¤Τإå󥯤ڤΥ
 */
STATIC void unlinkTimer(
	TIMER *m_timer,				// ڤΥޡ¤
	TIMER_HEAD *m_timerHead)	// ³西ޡإå
{
	AggregateList *aggregate = &m_timerHead->aggregate;
	TIMER *next = aggregate->refNext(aggregate, &m_timer->timerList);

	ASSERT(m_timer != NULL);

	if (next != NULL) {
		next->count += m_timer->count;
	}
	aggregate->removeEntry(aggregate, &m_timer->timerList);
}

/*
 *Գ
 *Թ
 * ֤вᤷޡ
 */
STATIC void getTimeoutTimer(
	const uint i_time,						// вms
	TIMER_HEAD *m_timerHead,				// ³西ޡإå
	AggregateList *m_aggregateUnlinkList)	// ޡ¤Τ³ñꥹȥإå
{
	uint time = i_time;
	AggregateList *aggregate = &m_timerHead->aggregate;
	TIMER *timer;

	for (timer = aggregate->refHead(aggregate); timer != NULL;) {
		TIMER *next = aggregate->refNext(aggregate, &timer->timerList);

		if (timer->count <= time) {
			aggregate->removeEntry(aggregate, &timer->timerList);
			time -= timer->count;
			timer->count = 0;
			m_aggregateUnlinkList->insertHead(m_aggregateUnlinkList, &timer->timerList);
		}
		else {
			timer->count -= time;
			break;
		}
		
		timer = next;
	}
}

/*
 *Թ
 * ޡ󥯤֤򸺤餹
 */
STATIC void subtractTimeFromTimerLink(
	const uint time,			// ms
	TIMER_HEAD *i_timerHead)	// ³西ޡإå
{
	AggregateList *aggregate = &i_timerHead->aggregate;
	TIMER *timer = aggregate->refHead(aggregate);
	
	ASSERT(timer != NULL);
	ASSERT(time <= timer->count);
	
	timer->count -= time;
}

/*
 *Ի
 * ޡκû֤
 * return : ûms
 */
STATIC uint getMinTimeFromTimerLink(
	TIMER_HEAD *i_timerHead)	// ³西ޡإå
{
	AggregateList *aggregate = &i_timerHead->aggregate;
	TIMER *timer = aggregate->refHead(aggregate);
	
	if (timer == NULL) {
		return 0;
	}
	else {
		return timer->count;
	}
}

/*
 * ޡϿե饰ꤹ
 */
STATIC INLINE void setRegist(
	TIMER *m_timer,
	TIMER_HEAD *timerHead)
{
	m_timer->timerHead = timerHead;
}

/*
 * ޡϿե饰
 */
STATIC INLINE void resetRegist(
	TIMER *m_timer)
{
	m_timer->timerHead = NULL;
}

/*
 * ޡꤹ
 */
STATIC void setTimerLink(
	const uint time,			// 
	TIMER *m_timer,				// ޡ¤
	TIMER_HEAD *m_timerHead)	// ³西ޡإå
{
	// ޡ󥯤Ͽ
	linkTimer(time, m_timer, m_timerHead);
	setRegist(m_timer, m_timerHead);
}

/*
 *ԳաԹ
 * ֤вᤷޡϥɥ¹Ԥ롣
 */
STATIC void doTimer(
	const uint i_time,			// в
	TIMER_HEAD *m_timerHead)	// ³إå
{
	AggregateList aggregateUnlilnkTimer;	// ñꥹ
	IteratorList iterator;
	int eflags;

	AggregateSingleListConstructor(&aggregateUnlilnkTimer);

	eflags = enterCli();
	enter_spinlock(&m_timerHead->lock);
	{
		// ॢȥޡμ
		getTimeoutTimer(i_time, m_timerHead, &aggregateUnlilnkTimer);
	}
	exit_spinlock(&m_timerHead->lock);
	exitCli(eflags);

	// ޡϥɥμ¹
	aggregateUnlilnkTimer.iterator(&aggregateUnlilnkTimer, &iterator);
	while (iterator.hasNext(&iterator) == BOOL_TRUE) {
		TIMER *timer = iterator.next(&iterator);

		timer->handle(timer->arg);
	}

	eflags = enterCli();
	enter_spinlock(&m_timerHead->lock);
	{
		uint setTime = UINT_MAX;
		TIMER *timer;

		while ((timer = aggregateUnlilnkTimer.getHead(&aggregateUnlilnkTimer)) != NULL) {
			// ³ޡͤк³
			if (0 < timer->init) {
				setTimerLink(timer->init, timer, m_timerHead);
				if (timer->init < setTime) {
					setTime = timer->init;
				}
			}
			else {
				resetRegist(timer);
			}

			// ǥХΥޡ֥å
			if (setTime < UINT_MAX) {
				m_timerHead->setTimer(setTime);
			}
		}
	}
	exit_spinlock(&m_timerHead->lock);
	exitCli(eflags);
}

//================================== PROTECT =============================================

/*
 *Ի
 * ޡ󥯤³Ƥ뤫
 * return : YES or NO
 */
STATIC INLINE int isRegistTimer(
	const TIMER *timer)
{
	return (timer->timerHead != NULL) ? YES : NO;
}

/*
 *ադδؿɬ delFromSchedule() θǻȤȡ⤷˥ॢȤȥץʤ롣
 *Թ
 * ޡ¤Τ򥿥ޡإåϿ
 * return : error number
 */
STATIC void registTimer(
	const uint i_time,	// ޡms
	TIMER  *m_timer)
{
	TIMER_HEAD *timerHead;
	int eflags;

	// ޡإå
	if (i_time < REAL_TIMER_RATE) {
		timerHead = &intervalTimer;
	}
	else {
		timerHead = &realTimer;
	}

	eflags = enterCli();
	enter_spinlock(&timerHead->lock);
	{
		// ޡȯԤޤǤθߤηв֤
		uint pastTime = timerHead->getPastTime();

		setTimerLink(i_time + pastTime, m_timer, timerHead);

		// ǥХޡ֤
		if (i_time + pastTime == getMinTimeFromTimerLink(timerHead)) {
			subtractTimeFromTimerLink(pastTime, timerHead);
			timerHead->setTimer(i_time);
		}
	}
	exit_spinlock(&timerHead->lock);
	exitCli(eflags);
}

/*
 * ޡ¤Τ򥿥ޡإå
 */
STATIC void unregistTimer(
	TIMER  *m_timer)
{
	TIMER_HEAD *timerHead = m_timer->timerHead;
	int eflags;

	ASSERT(m_timer != NULL);
	
	if (isRegistTimer(m_timer) == YES) {
		eflags = enterCli();
		enter_spinlock(&timerHead->lock);
		{
			if (isRegistTimer(m_timer) == YES) {
				unlinkTimer(m_timer, m_timer->timerHead);
				resetRegist(m_timer);
			}
		}
		exit_spinlock(&timerHead->lock);
		exitCli(eflags);
	}
}

/*
 *Ի
 * Ĥ֤
 * return : Ĥms
 */
STATIC uint getRemainTime(
	TIMER *i_timer)
{
	TIMER_HEAD *timerHead = i_timer->timerHead;
	uint time;
	int eflags;
	
	ASSERT(i_timer != NULL);

	time = 0;
	if (isRegistTimer(i_timer) == YES) {
		eflags = enterCli();
		enter_spinlock(&timerHead->lock);
		{
			if (isRegistTimer(i_timer) == YES) {
				AggregateList *aggregate = &timerHead->aggregate;
				IteratorList iterator;

				aggregate->iterator(aggregate, &iterator);
				for (;;) {
					TIMER *timer = iterator.next(&iterator);
					
					// i_timer¸ߤϤ
					ASSERT(timer != NULL);
					
					// Ĥ֤û
					time += timer->count;
					
					if (timer == i_timer) {
						break;
					}
				}
			}
		}
		exit_spinlock(&timerHead->lock);
		exitCli(eflags);
	}

	return time;
}

//================================== PUBLIC =============================================

/*
 *Գ
 * 󥿡Х륿ޡμ¹
 */
void doIntervalTimer(
	const uint time)		// вᤷ(ms)
{
	doTimer(time, &intervalTimer);
}

/*
 *Գ
 * ꥢ륯åޡμ¹
 */
void doRealTimer(
	const uint time)		// вᤷ(ms)
{
	doTimer(time, &realTimer);
}

/***************************************************************************************************
 * ޡ
 *
 * ñꥹ
 ***************************************************************************************************/

//================================== PRIVATE ============================================

enum{
	TIMER_MAX = 10,			// ޡ
};

static TIMER_OWNER systemTimer;	// ƥݻޡ

/*
 * ޡͭإåν
 */
STATIC INLINE void initOwnerHead(
	TIMER_OWNER *m_owner)
{
	AggregateSingleListConstructor(&m_owner->aggregate);
	m_owner->count = 0;
	m_owner->lock = 0;
}

/*
 *Ի
 * ͭإåϥɥõ
 * return : NOERR or -ENOENT
 */
STATIC int findTimerByHandle(
	TIMER_OWNER *i_owner,		// ͭإå
	void (*handle)(void*),
	TIMER **o_timer)
{
	AggregateList *aggregate = &i_owner->aggregate;
	IteratorList iterator;

	aggregate->iterator(aggregate, &iterator);
	while (iterator.hasNext(&iterator) == BOOL_TRUE) {
		TIMER *timer = iterator.next(&iterator);
		
		if (timer->handle == handle) {
			*o_timer = timer;
			return NOERR;
		}
	}

	return -ENOENT;
}

/*
 * ͭإåѥޡõ
 * return : NOERR or -ENOENT
 */
STATIC int findDelTimer(
	TIMER_OWNER *i_owner,		// ͭإå
	TIMER **o_timer)
{
	AggregateList *aggregate = &i_owner->aggregate;
	IteratorList iterator;

	aggregate->iterator(aggregate, &iterator);
	while (iterator.hasNext(&iterator) == BOOL_TRUE) {
		TIMER *timer = iterator.next(&iterator);
		
		if (isRegistTimer(timer) == NO) {
			*o_timer = timer;
			return NOERR;
		}
	}

	return -ENOENT;
}

/*
 * ޡƽ
 * return : error number
 */
STATIC int allocTimer(
	const uint initTime,	// ַв˺Ƥꤹ륿ޡ
	void (*handle)(void*),	// ַв˼¹Ԥ
	void *arg,				// ַв˼¹Ԥΰ
	TIMER_OWNER *m_owner,	// ޡͭ
	TIMER **o_timer)
{
	TIMER *timer;

	ASSERT(handle != NULL);

	// ޡõƤʤгƤ
	if (findDelTimer(m_owner, &timer) == -ENOENT) {
		AggregateList *aggregate = &m_owner->aggregate;

		if (TIMER_MAX <= m_owner->count) {
			WARNING(0);
			return -ENOMEM;
		}

		timer = kmalloc(sizeof(*timer));
		if (timer == NULL){
			return -ENOMEM;
		}
		listConstructor(&timer->ownedList, timer);
		aggregate->insertHead(aggregate, &timer->ownedList);
		++m_owner->count;
	}

	listConstructor(&timer->timerList, timer);
	timer->init = initTime;
	timer->handle = handle;
	timer->arg = arg;

	*o_timer = timer;

	return NOERR;
}

//================================== PROTECT ============================================

/*
 * ץޡν
 */
STATIC void initProcessTimer(PROC *proc)
{
	initOwnerHead(&proc->procTimer);
}

/*
 * ƥॿޡƤ
 * return : error number
 */
STATIC int allocSystemTimer(
	const uint i_time,		// ޡ
	const uint initTime,	// ַв˺Ƥꤹ륿ޡ
	void (*handle)(void*),	// ַв˼¹Ԥ
	void *arg,				// ַв˼¹Ԥΰ
	TIMER **o_timer)
{
	int error;

	enter_spinlock(&systemTimer.lock);
	{
		error = allocTimer(initTime, handle, arg, &systemTimer, o_timer);
	}
	exit_spinlock(&systemTimer.lock);
	
	if (error == NOERR) {
		registTimer(i_time, *o_timer);
		return NOERR;
	}
	else {
		return error;
	}
}

/*
 *Թ
 * ץޡƤ
 * return : error number
 */
STATIC int allocProcessTimer(
	const uint i_time,		// ޡ
	const uint initTime,	// ַв˺Ƥꤹ륿ޡ
	void (*handle)(void*),	// ַв˼¹Ԥ
	void *arg,				// ַв˼¹Ԥΰ
	TIMER **o_timer)
{
	int error = allocTimer(initTime, handle, arg, &get_current_task()->procTimer, o_timer);

	if (error == NOERR) {
		registTimer(i_time, *o_timer);
		return NOERR;
	}
	else {
		return error;
	}
}

/*
 * ޡ
 * return : error number
 */
STATIC void freeTimer(
	TIMER *m_timer)
{
	// Ͽ
	unregistTimer(m_timer);
	
	// ޡ¤Τϳʤ
}

/*
 *Թ
 * ץͭޡ
 */
STATIC void releaseProcessTimer(
	TIMER_OWNER *m_owner)
{
	AggregateList *aggregate = &m_owner->aggregate;
	TIMER *timer;

	while ((timer = aggregate->getHead(aggregate)) != NULL) {
		// Ͽ
		unregistTimer(timer);

		kfree(timer);
	}
}

/*
 *Գ
 * 顼ϥɥ
 */
STATIC void handleAlarm(void *proc)
{
	sendSignal(proc, SIGALRM);
}

/*
 *Գ
 * ϥɥ
 */
STATIC void handleWake(void *proc)
{
	forceSendSignal(proc, SIGWAKE, TASK_SIGNAL_WAIT);
}

/*
 * ϥɥ
 */
STATIC void handleDevWake(void *proc)
{
	forceSendSignal(proc, SIGWAKE, TASK_DEVICE_WAIT);
}

/*
 *Գ
 * 󥰥ϥɥ
 */
STATIC void handleTasking(
	void *dummy)		// ̤
{
	// åԤ
	setIntervalTasking(YES);
/********************************************************************/
	{
		static char value = '0';

		if ('9' < value){
			value = '0';
		}
		displayTop(60, &value, 1);
		++value;
	}
/********************************************************************/
}

//================================== PUBLIC ==============================================

/*
 * ǥХԤޡΥå
 * return : error number
 */
int setDevWakeTimer(
	const uint i_time,		// ޡ֡ms
	void **o_timer)
{
	ASSERT(0 < i_time);

	// ޡκ
	return allocProcessTimer(i_time, 0, handleDevWake, get_current_task(), (TIMER**) o_timer);
}

/*
 * ޡβ
 */
void delWakeTimer(
	void *i_timer)		// ޡ
{
	freeTimer(i_timer);
}

/*
 * 󥰤
 * return : error number
 */
int setTasking(
	const uint i_time)		// 󥰻֡ms
{
	TIMER *timer;
	
	ASSERT(0 < i_time);

	// ޡκ
	return allocSystemTimer(i_time, i_time, handleTasking, NULL, &timer);
}

/*
 * timeout --
 *	Execute a function after a specified length of time.
 *
 * untimeout --
 *	Cancel previous timeout function call.
 *
 * callout_handle_init --
 *	Initialize a handle so that using it with untimeout is benign.
 *
 *	See AT&T BCI Driver Reference Manual for specification.  This
 *	implementation differs from that one in that although an 
 *	identification value is returned from timeout, the original
 *	arguments to timeout as well as the identifier are used to
 *	identify entries for untimeout.
 *
 *Գ
 *ա¾ץƤФ礢
 */
void callout_handle_init(struct callout_handle *callout)
{
	callout->callout = NULL;
}

struct callout_handle timeout(
	timeout_t *ftn,
	void *arg, 
	int to_ticks)		// ms
{
	TIMER *timer;
	struct callout_handle callout;

	if (to_ticks == 0) {
		callout.callout = NULL;
		return callout;
	}

	if (allocSystemTimer(to_ticks, 0, ftn, arg, &timer) == NOERR) {
		callout.callout = timer;
	}
	else {
		printk("Failed timeout()!\n");
		callout.callout = NULL;
	}

	return callout;
}

/*
 *Գ
 */
void untimeout(
	timeout_t *ftn, 
	void *arg, 
	struct callout_handle callout)
{
	if (callout.callout == NULL){
		return;
	}

	freeTimer(callout.callout);
}

void callout_init(void *c)
{
	// Ȥꤢ⤷ʤ
}

void callout_reset(void *c, int to_ticks, void (*ftn)(void *), void *arg)
{
	// Ȥꤢ⤷ʤ
}

/*
 * return : 0 : ȥåcallout̵ or 1 : ȥåcalloutͭ
 */
int callout_stop(void *c)
{
	// Ȥꤢ⤷ʤ
	return 0;
}

/*
 *FreeBSD
 * General sleep call.  Suspends the current process until a wakeup is
 * performed on the specified identifier.  The process will then be made
 * runnable with the specified priority.  Sleeps at most timo/hz seconds
 * (0 means no timeout).  If pri includes PCATCH flag, signals are checked
 * before and after sleeping, else signals are not checked.  Returns 0 if
 * awakened, EWOULDBLOCK if the timeout expires.  If PCATCH is set and a
 * signal needs to be delivered, ERESTART is returned if the current system
 * call should be restarted if possible, and EINTR is returned if the system
 * call should be interrupted by the signal (return EINTR).
 */
int	tsleep(
	void *chan, 
	int pri, 
	const char *wmesg, 
	int timeout)
{
	struct buf *bp = chan;
	TIMER *timer;
	uint remainTime;
	int error;

	bp->proc = get_current_task();

	if (0 < timeout) {
		error = allocProcessTimer(timeout, 0, handleWake, get_current_task(), &timer);
		if (error != NOERR) {
			return 0;
		}
	}
	else if (timeout == 0) {
		// ॢȤʤ
		error = allocProcessTimer(UINT_MAX, UINT_MAX, handleWake, get_current_task(), &timer);
		if (error != NOERR){
			return 0;
		}
	}
	else {
		return 0;
	}
	sigWait();
	if ((pri & PCATCH) != 0) {
		if (isSigint() == TRUE) {
			return EINTR;
		}
	}

	// Ĥ֤
	remainTime = getRemainTime(timer);

	// 
	freeTimer(timer);
	
	if (remainTime == 0) {
		return EWOULDBLOCK;
	}
	else {
		return 0;
	}
}

int sys_sleep(uint seconds)
{
	TIMER *timer;
	uint remainTime;
	int error;

	if (seconds == 0){
		return 0;
	}

	error = allocProcessTimer(seconds * 1000, 0, handleWake, get_current_task(), &timer);
	if (error != NOERR){
		return seconds;
	}

	sigWait();

	// Ĥ֤
	remainTime = getRemainTime(timer);

	// 
	freeTimer(timer);

	return remainTime;
}

int sys_alarm(uint seconds)
{
	TIMER *timer;

	// 顼ॿޡˤв
	if (findTimerByHandle(&get_current_task()->procTimer, handleAlarm, &timer) == NOERR) {
		freeTimer(timer);
	}

	// ץޡ
	if (0 < seconds){
		return allocProcessTimer(seconds * 1000, 0, handleAlarm, get_current_task(), &timer);
	}
	else {
		return 0;
	}
}

/***************************************************************************************************
 *
 * ץ󥿡Х륿ޡ
 *
 ***************************************************************************************************/

//================================== PRIVATE ============================================

/*
 * ץ󥿡Х륿ޡν
 */
STATIC void initProcItimer(
	PROC *proc)
{
	proc->itimer[ITIMER_VIRTUAL].init = 0;
	proc->itimer[ITIMER_VIRTUAL].current = 0;
	proc->itimer[ITIMER_PROF].init = 0;
	proc->itimer[ITIMER_PROF].current = 0;
}

/*
 *struct timevalͤ0ʾǤ뤳
 * struct timeval ms֤
 * return : mili seconds
 */
STATIC INLINE uint getMiliSecondsFromTimeval(
	const struct timeval *i_timeval)
{
	int miliSeconds;
	
	miliSeconds = i_timeval->tv_sec * 1000;
	miliSeconds += i_timeval->tv_usec / 1000;
	
	return miliSeconds;
}

/*
 * ms֤struct timeval򥻥åȤ
 */
STATIC INLINE void setTimevalFromMiliSeconds(
	const uint i_miliSeconds,
	struct timeval *m_timeval)
{
	m_timeval->tv_sec = i_miliSeconds / 1000;
	m_timeval->tv_usec = (i_miliSeconds % 1000) * 1000;
}

/*
 * struct timevalCPUå
 */
STATIC INLINE uint64 getClocksFromTimeval(
	const struct timeval *i_timeval)
{
	uint64 clocks;
	
	clocks = calcClockFromMicroSecond(i_timeval->tv_sec * 1000000);
	clocks += calcClockFromMicroSecond(i_timeval->tv_usec);

	return clocks;
}

/*
 * CPUåstruct timeval򥻥åȤ
 */
STATIC INLINE void setTimevalFromClocks(
	const uint64 i_clocks,
	struct timeval *m_timeval)
{
	uint64 microSeconds = calcMicroSecondFromClock(i_clocks);

	m_timeval->tv_sec = microSeconds / 1000000;
	m_timeval->tv_usec = microSeconds % 1000000;
}

/*
 * itimeritimervalͤꤹ
 */
STATIC void setItimervalFromItimer(
	const int i_which,
	struct itimerval *m_itimerval)
{
	ITIMER *itimer = &get_current_task()->itimer[i_which];
	
	setTimevalFromClocks(itimer->init, &m_itimerval->it_interval);
	setTimevalFromClocks(itimer->current, &m_itimerval->it_value);
}

/*
 * itimervalitimerͤꤹ
 */
STATIC void setItimerFromItimerval(
	const int i_which,
	const struct itimerval *i_itimerval)
{
	ITIMER *itimer = &get_current_task()->itimer[i_which];
	
	itimer->init = getClocksFromTimeval(&i_itimerval->it_interval);
	itimer->current = getClocksFromTimeval(&i_itimerval->it_value);
}

/*
 * 󥿡Х륿ޡ¹Ԥ
 */
STATIC void doItimer(
	const int i_which,
	const int64 clocks,		// вCPUå
	const int signal)		// ޡв륷ʥ
{
	ITIMER *itimer = &get_current_task()->itimer[i_which];
	
	if (0 < itimer->current) {
		itimer->current -= clocks;
		if (itimer->current <= 0) {
			sendSignal(get_current_task(), signal);
			itimer->current = itimer->init;
		}
	}
}

//================================== PUBLIC =============================================

/*
 * ƥв֤ˤ륤󥿡Х륿ޡ
 */
void doSystemItimer(
	const int64 clocks)		// вCPUå
{
	doItimer(ITIMER_PROF, clocks, SIGPROF);
}

/*
 * 桼в֤ˤ륤󥿡Х륿ޡ
 */
void doUserItimer(
	const int64 clocks)		// вCPUå
{
	doItimer(ITIMER_PROF, clocks, SIGPROF);
	doItimer(ITIMER_VIRTUAL, clocks, SIGVTALRM);
}

int sys_getitimer(const int i_which, struct itimerval *m_itimerval)
{
	if (checkMem(m_itimerval, sizeof(*m_itimerval)) == -1){
		return -EFAULT;
	}

	switch (i_which) {
	case ITIMER_VIRTUAL:
		// 
	case ITIMER_PROF:
		setItimervalFromItimer(i_which, m_itimerval);
		break;
	case ITIMER_REAL: {
		TIMER *timer;

		if (findTimerByHandle(&get_current_task()->procTimer, handleAlarm, &timer) == NOERR) {
			// ߤλ֤
			setTimevalFromMiliSeconds(getInitTime(timer), &m_itimerval->it_interval);
			setTimevalFromMiliSeconds(getRemainTime(timer), &m_itimerval->it_value);
		}
		break;
	}
	default:
		return -EFAULT;
	}

	return 0;
}

int sys_setitimer(
	const int i_which,
	const struct itimerval *i_itimerval,
	struct itimerval *m_itimerval)
{
	if (checkMem(i_itimerval, sizeof(*i_itimerval)) == -1){
		return -EFAULT;
	}
	if (m_itimerval != NULL){
		if (checkMem(m_itimerval, sizeof(*m_itimerval)) == -1){
			return -EFAULT;
		}
	}
	if ((i_itimerval->it_interval.tv_sec < 0) || 
		(i_itimerval->it_interval.tv_usec < 0) ||
		(i_itimerval->it_value.tv_sec < 0) || 
		(i_itimerval->it_value.tv_usec < 0)) {
		return -EINVAL;
	}

	switch (i_which) {
	case ITIMER_VIRTUAL:
		// 
	case ITIMER_PROF:
		if (m_itimerval != NULL) {
			setItimervalFromItimer(i_which, m_itimerval);
		}
		setItimerFromItimerval(i_which, i_itimerval);
		break;
	case ITIMER_REAL: {
		TIMER *timer;
		uint time;
		uint initTime;

		// 顼ॿޡˤв
		if (findTimerByHandle(&get_current_task()->procTimer, handleAlarm, &timer) == NOERR) {
			if (m_itimerval != NULL) {
				setTimevalFromMiliSeconds(getInitTime(timer), &m_itimerval->it_interval);
				setTimevalFromMiliSeconds(getRemainTime(timer), &m_itimerval->it_value);
			}
			freeTimer(timer);
		}

		// ץޡ
		time = getMiliSecondsFromTimeval(&i_itimerval->it_value);
		initTime = getMiliSecondsFromTimeval(&i_itimerval->it_interval);
		if (0 < time){
			allocProcessTimer(time, initTime, handleAlarm, get_current_task(), &timer);
		}

		break;
	}
	default:
		return -EFAULT;
	}

	return 0;
}

/***************************************************************************************************
 * ¾ޡ
 ***************************************************************************************************/

//================================== PUBLIC =============================================

/*
 * micro seccond ޡ
 */
void micro_timer(
	uint count)		// ¬(s)
{
	uint64 end;

	end = rdtsc();
	end += calcClockFromMicroSecond(count);
	while (rdtsc() < end);
}

/*
 * mili seccond timer
 */
void mili_timer(
	uint count)		// ¬(ms)
{
	uint64 end;

	end = rdtsc();
	end += calcClockFromMiliSecond(count);
	while (rdtsc() < end);
}

/***************************************************************************************************
 * 
 ***************************************************************************************************/

/*
 * ޡν
 *ե֡Ȼ˸ƤӽФ
 */
void initTimer(void)
{
	// ƥॿޡإåν
	initTimerHead();

	// ƥॿޡͭإåν
	initOwnerHead(&systemTimer);
}

/*
 * ץޡν
 *եץ˸ƤӽФ
 */
void initProcTimer(void *i_proc)
{
	initProcessTimer(i_proc);
	initProcItimer(i_proc);
}

/*
 * ץޡγ
 */
void releaseProcTimer()
{
	releaseProcessTimer(&get_current_task()->procTimer);
}

/*
 *Գ
 */
void sys_wake(void *proc)
{
	forceSendSignal(proc, SIGWAKE, TASK_SIGNAL_WAIT);
}
